/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CalendarEventDelegate.h"

#include "CalendarEventModel.h"
#include "CalendarEvent.h"

#include <QAbstractTextDocumentLayout>
#include <QApplication>
#include <QStyleOptionViewItem>
#include <QPainter>

namespace Calendar
{

    //_________________________________________________________
    CalendarEventDelegate::CalendarEventDelegate( QObject *parent ):
        QAbstractItemDelegate( parent ),
        Counter( "CalendarEventDelegate" )
    { Debug::Throw( "CalendarEventDelegate::CalendarEventDelegate.\n" ); }

    //_________________________________________________________
    void CalendarEventDelegate::paint( QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index ) const
    {

        // check index
        if ( !index.isValid() ) return;

        // cast model
        const CalendarEventModel& model( *static_cast<const CalendarEventModel*>( index.model() ) );
        if( !model.contains( index ) ) return;

        // selection
        {
            QStyleOptionViewItemV4 viewOption(option);
            viewOption.showDecorationSelected = true;
            QStyle *style = viewOption.widget ? viewOption.widget->style() : QApplication::style();
            style->drawPrimitive(QStyle::PE_PanelItemViewItem, &viewOption, painter, viewOption.widget);
        }

        // get calendar event
        const CalendarEvent& event( model.get( index ) );

        // create layout
        QTextDocument document;
        document.setDefaultFont( option.font );
        QTextCursor cursor( &document );

        // modify color
        QTextCharFormat charFormat( cursor.charFormat() );
        QPalette::ColorGroup colorGroup = option.state & QStyle::State_Enabled ? QPalette::Normal : QPalette::Disabled;
        if( colorGroup == QPalette::Normal && !(option.state & QStyle::State_Active) )
        { colorGroup = QPalette::Inactive; }

        QColor foreground;

        if( option.state & QStyle::State_Selected ) foreground = option.palette.color( colorGroup, QPalette::HighlightedText );
        else foreground = option.palette.color( colorGroup, QPalette::Text );

        charFormat.setForeground( foreground );
        cursor.setCharFormat( charFormat );
        event.toLayout( cursor );

        painter->save();
        painter->translate( option.rect.topLeft() );
        painter->setClipRect( option.rect.translated( -option.rect.topLeft() ) );
        document.drawContents( painter );
        painter->restore();

        // focus
        _drawFocus( painter, option, option.rect );

    }

    //_________________________________________________________
    QSize CalendarEventDelegate::sizeHint( const QStyleOptionViewItem &option, const QModelIndex &index ) const
    {

        if ( !index.isValid() ) return QSize();

        // cast model
        const CalendarEventModel& model( *static_cast<const CalendarEventModel*>( index.model() ) );
        if( !model.contains( index ) ) return QSize();

        // get calendar event
        CalendarEvent event( model.get( index ) );

        // create document and layout text
        QTextDocument document;
        document.setDefaultFont( option.font );
        QTextCursor cursor( &document );
        event.toLayout( cursor );

        // size
        return QSize( document.idealWidth(), document.size().height() );

    }

    //_________________________________________________________
    void CalendarEventDelegate::_drawFocus( QPainter *painter, const QStyleOptionViewItem &option, const QRect &rect ) const
    {
        if (option.state & QStyle::State_HasFocus)
        {

            QStyleOptionFocusRect o;
            o.QStyleOption::operator=(option);
            o.rect = rect;
            o.state |= QStyle::State_KeyboardFocusChange;
            QPalette::ColorGroup colorGroup = (option.state & QStyle::State_Enabled) ? QPalette::Normal : QPalette::Disabled;
            o.backgroundColor = option.palette.color( colorGroup, (option.state & QStyle::State_Selected) ? QPalette::Highlight : QPalette::Background );

            QApplication::style()->drawPrimitive( QStyle::PE_FrameFocusRect, &o, painter );

        }

    }

}
