/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "CalendarWidget.h"

#include "AdjustTimeDialog.h"
#include "BaseContextMenu.h"
#include "CalendarConfigurationDialog.h"
#include "CalendarEventDialog.h"
#include "CalendarEventSmtpClient.h"
#include "CalendarHeader.h"
#include "CalendarToolTipWidget.h"
#include "Color.h"
#include "Debug.h"
#include "EditCalendarEventDialog.h"
#include "Effects.h"
#include "InformationDialog.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "IconSize.h"
#include "QuestionDialog.h"
#include "SelectMonthDialog.h"
#include "ToolTipWidget.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QDateTime>
#include <QListView>
#include <QProcess>
#include <QPainter>
#include <QStyleOptionViewItemV4>
#include <QToolTip>
#include <cmath>

namespace Calendar {

    //_______________________________________________
    CalendarWidget::CalendarWidget( QWidget* parent ):
        MainWindow( parent ),
        nColumns_( 7 ),
        nRows_( 7 ),
        cellWidth_( 0 ),
        cellHeight_( 0 )
    {

        Debug::Throw( "CalendarWidget::CalendarWidget.\n" );
        setWindowTitle( Util::windowTitle( tr( "Calendar" ) ) );
        setFocusPolicy( Qt::StrongFocus );

        // some styles require an item view passed to painting method to have proper selection rendered in items
        itemView_ = new QListView( this );
        itemView_->hide();

        // layout
        QVBoxLayout* layout = new QVBoxLayout;
        layout->setSpacing(0);
        layout->setMargin(0);
        setLayout( layout );
        layout->addWidget( header_ = new CalendarHeader( this ) );
        layout->addStretch( 1 );

        // actions
        _installActions();
        _configurationAction().setText( tr( "Transparent Calendar Settings..." ) );

        // header connections
        connect( &header_->previousButton(), SIGNAL(clicked()), SLOT(_selectPreviousMonth()) );
        connect( &header_->nextButton(), SIGNAL(clicked()), SLOT(_selectNextMonth()) );

        connect( &inverseColorsAction(), SIGNAL(toggled( bool )), SLOT(_updateHeaderColors()) );

        // update configuration
        _updateConfiguration();

        // read events
        events_.read( File( XmlOptions::get().raw( "RC_FILE" ) ) );
        _setDate( QDate::currentDate() );
        _checkNotifications();

        // restart time
        _restartTimer();

        setFocus();

    }

    //______________________________________________________________________
    void CalendarWidget::mousePressEvent( QMouseEvent* event )
    {

        Debug::Throw( "CalendarWidget::mousePressEvent.\n" );
        MainWindow::mousePressEvent( event );
        if( _button() == Qt::LeftButton && _lockPositionAction().isChecked() )
        {

            // first clear
            if( !( event->modifiers() & Qt::ShiftModifier ) ) _clearSelectedDays();
            _updateSelection( event->pos() - _margins().topLeft() - QPoint( 0, header_->contentsRect().height() ) );

        }

    }

    //______________________________________________________________________
    void CalendarWidget::mouseMoveEvent( QMouseEvent* event )
    {

        MainWindow::mouseMoveEvent( event );
        if( _button() == Qt::LeftButton && _lockPositionAction().isChecked() )
        {

            CalendarEvent::SelectionMode mode( ( event->modifiers() & Qt::ShiftModifier ) ? CalendarEvent::SelectionBox:CalendarEvent::SelectionRange );
            _updateSelection( event->pos() - _margins().topLeft() - QPoint( 0, header_->contentsRect().height() ), mode );

        }

        return;

    }


    //________________________________________________
    void CalendarWidget::contextMenuEvent( QContextMenuEvent* event )
    {

        Debug::Throw( "CalendarWidget::contextMenuEvent.\n" );

        bool hasCurrentEvent( false );
        const QDate currentDate( Cell::findCell( cells_, event->pos() - _margins().topLeft() - QPoint( 0, header_->contentsRect().height() ) ).date() );
        if( currentDate.isValid() )
        {

            // check if there is a matching current event
            CalendarEvent::List::const_iterator iter;
            if( ( iter = std::find_if( _currentEvents().begin(), _currentEvents().end(), CalendarEvent::ContainsFTor( currentDate ) ) ) != _currentEvents().end() )
            {

                hasCurrentEvent = true;
                currentEvent_ = *iter;

            }

            // find matching selection, if any
            if( ( iter = std::find_if( _selectedDays().begin(), _selectedDays().end(), CalendarEvent::ContainsFTor( currentDate ) ) ) != _selectedDays().end() )
            {

                selectedEvent_ = *iter;

            } else {

                selectedEvent_ = CalendarEvent();
                selectedEvent_.setRange( currentDate );

            }

        } else {

            const QDate today( QDate::currentDate() );
            selectedEvent_ = CalendarEvent();
            selectedEvent_.setBegin( QDateTime( today ) );
            selectedEvent_.setEnd( QDateTime( today ) );

        }

        editEventAction_->setEnabled( hasCurrentEvent );
        removeEventAction_->setEnabled( hasCurrentEvent );

        MainWindow::contextMenuEvent( event );

    }

    //______________________________________________________________________
    void CalendarWidget::timerEvent( QTimerEvent* event )
    {

        Debug::Throw() << "CalendarWidget::timerEvent - id: " << event->timerId() << endl;

        // local implementation
        if( event->timerId() == _geometryTimer().timerId() )
        {

            /*
            the geometry timer is not stopped here,
            because this is done in MainWindow::timerEvent
            */
            _updateGeometry();
            _updateCurrentPathList();
            _updateSelectedPathList();
            update();

        } else if( event->timerId() == timer_.timerId() ) {

            // update month if needed
            const QDate today( QDate::currentDate() );
            const QDate previous( today.addDays( -1 ) );
            if( date_.isNull() || ( date_.year() == previous.year() && date_.month() == previous.month() ) )
            { _setDate( today ); }

            // notifications
            _checkNotifications();

            // update display
            update();

            // restart timer
            timer_.stop();
            _restartTimer();

        }

        // base class method
        MainWindow::timerEvent( event );

    }

    //______________________________________________________________________
    void CalendarWidget::_installContextMenuActions( BaseContextMenu* menu )
    {
        Debug::Throw( "CalendarWidget::_installContextMenuActions.\n" );

        // get menu from parent
        MainWindow::_installContextMenuActions( menu );

        // customize
        menu->insertSeparator( &_menuSeparator() );
        menu->insertAction( &_menuSeparator(), previousYearAction_ );
        menu->insertAction( &_menuSeparator(), previousMonthAction_ );
        menu->insertAction( &_menuSeparator(), currentMonthAction_ );
        menu->insertAction( &_menuSeparator(), nextMonthAction_ );
        menu->insertAction( &_menuSeparator(), nextYearAction_ );
        menu->insertSeparator( &_menuSeparator() );
        menu->insertAction( &_menuSeparator(), selectMonthAction_ );
        menu->insertAction( &_menuSeparator(), adjustTimeAction_ );
        menu->insertSeparator( &_menuSeparator() );
        menu->insertAction( &_menuSeparator(), addEventAction_ );
        menu->insertAction( &_menuSeparator(), removeEventAction_ );
        menu->insertAction( &_menuSeparator(), editEventAction_ );
        menu->insertAction( &_menuSeparator(), configureEventsAction_ );

    }

    //__________________________________________________________
    BaseToolTipWidget* CalendarWidget::_createToolTipWidget()
    {
        CalendarToolTipWidget* widget = new CalendarToolTipWidget( this );
        widget->setPixmap( windowIcon().pixmap( IconSize::get( IconSize::Maximum ) ) );
        return widget;

    }

    //________________________________________________
    void CalendarWidget::_configuration()
    {

        Debug::Throw( "CalendarWidget::_configuration.\n" );

        CalendarConfigurationDialog dialog;
        _configureFromDialog( dialog );

    }

    //________________________________________________
    void CalendarWidget::_adjustTime()
    {

        Debug::Throw( "CalendarWidget::_adjustTime.\n" );

        const bool alwaysUseCommand( XmlOptions::get().get<bool>( "ALWAYS_USE_CMD" ) );
        QString command( XmlOptions::get().contains( "ADJUST_TIME_CMD" ) ? QString( XmlOptions::get().raw( "ADJUST_TIME_CMD" ) ).trimmed():"" );

        if( !( alwaysUseCommand && !command.isEmpty() ) )
        {

            // create dialog
            AdjustTimeDialog dialog( this );
            dialog.setCommand( command );
            dialog.setAlwaysUse( alwaysUseCommand );
            if( dialog.exec() == QDialog::Rejected ) return;

            // retrieve command
            command = dialog.command().trimmed();

            // check if valid
            if( command.isEmpty() )
            {
                InformationDialog( this, tr( "No command set to adjust time. <Adjust Time> canceled." ) ).exec();
                return;
            }

            XmlOptions::get().set<bool>( "ALWAYS_USE_CMD", dialog.alwaysUse() );

        }

        // try run
        int exitCode( QProcess::execute( command ) );
        if( exitCode != 0 )
        {
            const QString buffer = QString( tr( "Command '%1' excited with code %2" ) ).arg( command ).arg( exitCode );
            InformationDialog( this, buffer ).exec();

            XmlOptions::get().set<bool>( "ALWAYS_USE_CMD", false );

        } else {
            XmlOptions::get().setRaw( "ADJUST_TIME_CMD", command );
            XmlOptions::write();
        }

        update();

        return;

    }

    //________________________________________________
    void CalendarWidget::_editEvent()
    {
        Debug::Throw( "CalendarWidget::_editEvent.\n" );

        EditCalendarEventDialog dialog( this, currentEvent_ );
        if( !dialog.centerOnDesktop().exec() ) return;

        // retrieve event and check validity
        CalendarEvent event( dialog.event() );
        QString error( event.validate() );
        if( !error.isEmpty() ) InformationDialog( this, error ).centerOnDesktop().exec();

        events_.replace( currentEvent_, event );
        events_.write();

        _updateCurrentEvents();
        _updateCurrentPathList();
        update();

    }

    //________________________________________________
    void CalendarWidget::_removeEvent()
    {
        Debug::Throw( "CalendarWidget::_removeEvents.\n" );

        QString buffer;
        if( !currentEvent_.description().isEmpty() ) buffer = QString( tr( "Remove the following event ?\n%1" ) ).arg( currentEvent_.description() );
        else buffer = QString( tr( "Remove the selected event ?" ) );
        if( !QuestionDialog( this, buffer ).centerOnWidget( window() ).exec() ) return;
        events_.remove( currentEvent_ );
        events_.write();

        _updateCurrentEvents();
        _updateCurrentPathList();
        update();

    }

    //________________________________________________
    void CalendarWidget::_addEvent()
    {
        Debug::Throw( "CalendarWidget::_addEvent.\n" );

        EditCalendarEventDialog dialog( this, selectedEvent_ );
        if( !dialog.centerOnWidget( window() ).exec() ) return;

        CalendarEvent event( dialog.event() );
        QString error( event.validate() );
        if( !error.isEmpty() ) InformationDialog( this, error ).centerOnWidget( window() ).exec();

        events_.add( event );
        events_.write();

        // update events
        _updateCurrentEvents();
        _updateCurrentPathList();

        // update selection
        _clearSelectedDays();
        _updateSelectedPathList();

        // update
        update();


    }

    //________________________________________________
    void CalendarWidget::_configureEvents()
    {
        Debug::Throw( "CalendarWidget::_configureEvents.\n" );

        // create dialog
        CalendarEventDialog dialog( this, events_.events() );
        if( !dialog.centerOnDesktop().exec() ) return;

        // update
        events_.setEvents( dialog.list() );
        events_.write();

        // update events and redraw
        _updateCurrentEvents();
        _updateCurrentPathList();
        update();

        return;

    }

    //________________________________________________
    void CalendarWidget::_toggleShowPastEvents( bool value )
    {
        XmlOptions::get().set<bool>( "SHOW_PAST_TAGS", value );
        _updateCurrentEvents();
        _updateCurrentPathList();
        update();
    }

    //________________________________________________
    void CalendarWidget::_toggleShowWeekNumbers( bool value )
    {
        Debug::Throw() << "CalendarWidget::_toggleShowWeekNumbers - " << value << endl;
        XmlOptions::get().set<bool>( "SHOW_WEEK_NUMBERS", value );
        _updateGeometry();
        _updateCurrentPathList();
        _updateSelectedPathList();
        update();
    }

    //________________________________________________
    void CalendarWidget::_toggleShowInvalidDays( bool value )
    {
        Debug::Throw() << "CalendarWidget::_toggleShowInvalidDays - " << value << endl;
        XmlOptions::get().set<bool>( "SHOW_INVALID_DAYS", value );
        _updateCurrentEvents();
        _updateCurrentPathList();
        _updateSelectedPathList();
        update();
    }

    //________________________________________________
    void CalendarWidget::_eventNotificationSent( const CalendarEvent& event )
    {
        Debug::Throw(0, "CalendarWidget::_eventNotificationSent.\n" );

        // find matching persistent event
        CalendarEvent::List& events( events_.events() );
        const int index( events.indexOf( event ) );
        if( index < 0 ) return;

        CalendarEvent& persistentEvent( events[index] );
        persistentEvent.setFlag( CalendarEvent::Notified, true );
        events_.write();
        _updateCurrentEvents();

    }

    //________________________________________________
    void CalendarWidget::_updateConfiguration()
    {

        Debug::Throw( "CalendarWidget::_updateConfiguration.\n" );

        // colors
        currentDayColor_ = XmlOptions::get().get<Base::Color>( "CURRENT_DAY_COLOR" );

        showPastEventsAction_->setChecked( XmlOptions::get().get<bool>( "SHOW_PAST_TAGS" ) );
        showWeekNumbersAction_->setChecked( XmlOptions::get().get<bool>( "SHOW_WEEK_NUMBERS" ) );
        showInvalidDaysAction_->setChecked( XmlOptions::get().get<bool>( "SHOW_INVALID_DAYS" ) );

        // base class update
        MainWindow::_updateConfiguration();
        _updateGeometry();
        _updateCurrentPathList();
        _updateSelectedPathList();

        // header
        QMetaObject::invokeMethod( this, "_updateHeaderColors", Qt::QueuedConnection );

        // repaint
        update();

    }

    //______________________________________________________________________
    void CalendarWidget::_updateHeaderColors()
    {
        // shadow
        header_->label().setShadow( shadowColor(), shadowOffset() );
        header_->label().setContentsMargins( shadowOffset(), shadowOffset(), shadowOffset(), shadowOffset() );

        // foreground
        QPalette palette( header_->palette() );
        palette.setColor( QPalette::Foreground, foregroundColor() );
        header_->setPalette( palette );
    }

    //______________________________________________________________________
    void CalendarWidget::_selectMonthFromDialog()
    {

        Debug::Throw( "CalendarWidget::_selectMonthFromDialog.\n" );
        SelectMonthDialog dialog( this, date_ );
        if( !dialog.centerOnPointer().exec() ) return;
        if( _setDate( dialog.date() ) ) update();

    }

    //______________________________________________________________________
    void CalendarWidget::_installActions()
    {

        Debug::Throw( "CalendarWidget::_installActions.\n" );

        addAction( previousYearAction_ = new QAction( IconEngine::get().get( IconNames::PreviousYear ), tr( "Previous Year" ), this ) );
        previousYearAction_->setShortcut( Qt::Key_Up );
        connect( previousYearAction_, SIGNAL(triggered()), SLOT(_selectPreviousYear()) );

        addAction( nextYearAction_ = new QAction( IconEngine::get().get( IconNames::NextYear ), tr( "Next Year" ), this ) );
        nextYearAction_->setShortcut( Qt::Key_Down );
        connect( nextYearAction_, SIGNAL(triggered()), SLOT(_selectNextYear()) );

        addAction( previousMonthAction_ = new QAction( IconEngine::get().get( IconNames::Previous ), tr( "Previous Month" ), this ) );
        previousMonthAction_->setShortcut( Qt::Key_Left );
        connect( previousMonthAction_, SIGNAL(triggered()), SLOT(_selectPreviousMonth()) );

        addAction( nextMonthAction_ = new QAction( IconEngine::get().get( IconNames::Next ), tr( "Next Month" ), this ) );
        nextMonthAction_->setShortcut( Qt::Key_Right );
        connect( nextMonthAction_, SIGNAL(triggered()), SLOT(_selectNextMonth()) );

        addAction( currentMonthAction_ = new QAction( IconEngine::get().get( IconNames::DialogAccept ), tr( "Today" ), this ) );
        currentMonthAction_->setShortcut( Qt::Key_Home );
        connect( currentMonthAction_, SIGNAL(triggered()), SLOT(_selectCurrentMonth()) );

        addAction( selectMonthAction_ = new QAction( IconEngine::get().get( IconNames::Edit ), tr( "Select Month..." ), this ) );
        connect( selectMonthAction_, SIGNAL(triggered()), SLOT(_selectMonthFromDialog()) );

        addAction( adjustTimeAction_ = new QAction( tr( "Adjust Time..." ), this ) );
        adjustTimeAction_->setToolTip( tr( "Run external command to adjust system date and time" ) );
        connect( adjustTimeAction_, SIGNAL(triggered()), SLOT(_adjustTime()) );

        addAction( showWeekNumbersAction_ = new QAction( tr( "Show Week Numbers" ), this ) );
        showWeekNumbersAction_->setCheckable( true );
        showWeekNumbersAction_->setChecked( true );
        connect( showWeekNumbersAction_, SIGNAL(toggled(bool)), SLOT(_toggleShowWeekNumbers(bool)) );

        addAction( showInvalidDaysAction_ = new QAction( tr( "Show Days from Previous/Next Month" ), this ) );
        showInvalidDaysAction_->setCheckable( true );
        showInvalidDaysAction_->setChecked( true );
        connect( showInvalidDaysAction_, SIGNAL(toggled(bool)), SLOT(_toggleShowInvalidDays(bool)) );

        // events
        addAction( addEventAction_ = new QAction( IconEngine::get().get( IconNames::AddEvent ), tr( "Add Event..." ), this ) );
        connect( addEventAction_, SIGNAL(triggered()), SLOT(_addEvent()) );

        addAction( removeEventAction_ = new QAction( IconEngine::get().get( IconNames::RemoveEvent ), tr( "Remove Event" ), this ) );
        connect( removeEventAction_, SIGNAL(triggered()), SLOT(_removeEvent()) );

        addAction( editEventAction_ = new QAction( IconEngine::get().get( IconNames::Edit ), tr( "Edit Event" ), this ) );
        connect( editEventAction_, SIGNAL(triggered()), SLOT(_editEvent()) );

        addAction( configureEventsAction_ = new QAction( IconEngine::get().get( IconNames::Configure ), tr( "Configure Events..." ), this ) );
        connect( configureEventsAction_, SIGNAL(triggered()), SLOT(_configureEvents()) );

        addAction( showPastEventsAction_ = new QAction( tr( "Show Past Events" ), this ) );
        showPastEventsAction_->setCheckable( true );
        showPastEventsAction_->setChecked( false );
        connect( showPastEventsAction_, SIGNAL(toggled(bool)), SLOT(_toggleShowPastEvents(bool)) );

    }

    //______________________________________________________________________
    void CalendarWidget::_updateCurrentEvents()
    {
        Debug::Throw( "CalendarWidget::_updateCurrentEvents.\n" );

        // check cells
        if( cells_.empty() ) return;

        currentEvents_.clear();

        // starting time
        QDate first( showInvalidDaysAction_->isChecked() ? cells_.front().date() : QDate( date_.year(), date_.month(), 1 ) );
        if( !showPastEventsAction_->isChecked() )
        {
            const QDate today( QDate::currentDate() );
            if( first < today ) first = today;
        }

        // ending time
        const QDate second( showInvalidDaysAction_->isChecked() ? cells_.back().date() : QDate( date_.year(), date_.month(), date_.daysInMonth() ) );
        for( auto&& event:events_.events() )
        {
            event.generate( first, second );
            if( event.ranges().empty() ) continue;
            currentEvents_ << event;
        }

    }

    //______________________________________________________________________
    void CalendarWidget::_updateGeometry()
    {

        Debug::Throw( "CalendarWidget::_updateGeometry.\n" );

        if( date_.isNull() ) return;

        // first day
        QDate first( date_.year(), date_.month(), 1 );
        const int maxDay( first.daysInMonth() );
        first = first.addDays( 1-first.dayOfWeek() );

        // last day
        QDate last( date_.year(), date_.month(), maxDay );
        last = last.addDays( 7-last.dayOfWeek() );

        // number of columns
        if( showWeekNumbersAction_->isChecked() )
        {

            nColumns_ = 8;
            cellWidth_ = qreal( width() - 3 - _margins().width() )/nColumns_;
            header_->setContentsMargins( _margins().left()*1.125 + cellWidth_, _margins().top()+2, _margins().right(), 2 );

        } else {

            nColumns_ = 7;
            cellWidth_ = qreal( width() - _margins().width() )/nColumns_;
            header_->setContentsMargins( _margins().left(), _margins().top()+2, _margins().right(), 2 );

        }

        // number of rows
        nRows_ = 2+first.daysTo( last )/7;
        cellHeight_ = qreal( height() - 2 - header_->contentsRect().height() - _margins().height() )/nRows_;

        // one should store dates and Rect in an array here.
        // this would make the painting much faster
        cells_.clear();
        weekCells_.clear();
        weekDayCells_.clear();

        // define day cells
        int column(0);
        int row(0);

        QDate current( first );
        for( int day = 0; current <= last; current = current.addDays(1), ++day )
        {

            // QPointF origin( _cellWidth()*column, cellHeight_*(1 + row) );
            QPointF origin( _cellWidth()*column, cellHeight_*row );

            // add week cell
            if( showWeekNumbersAction_->isChecked() && column == 0 && row <= _nRows()-2 )
            { weekCells_ << Cell( current, QRectF( origin + QPointF( 0, cellHeight_ + 2 ), _cellSize() ) ); }

            // add offset
            if( showWeekNumbersAction_->isChecked() )
            { origin += QPointF( _cellWidth() + 3, 0 ); }

            // add weekDay cell
            if( row == 0 )
            { weekDayCells_ << Cell( current, QRectF( origin, _cellSize() ) ); }

            origin += QPointF( 0, cellHeight_ + 2 );

            // add cell to list
            Cell cell( current, QRectF( origin, _cellSize() ) );
            cell.setEnabled( current.month() == date_.month() );

            cells_ << cell;

            // increment column
            column++;
            if( (  showWeekNumbersAction_->isChecked() && column >= _nColumns()-1 ) || column >= _nColumns() )
            {
                column = 0;
                row++;
            }

        }

    }

    //______________________________________________________________________
    Path::List CalendarWidget::_pathList( const CalendarEvent::List& events ) const
    {

        Debug::Throw( "CalendarWidget::_pathList.\n" );

        if( date_.isNull() ) return Path::List();

        QDate first( date_.year(), date_.month(), 1 );
        QDate last( date_.year(), date_.month(), date_.daysInMonth() );

        Path::List pathList;
        for( const auto& event:events )
        {

            // store rects
            using RectangleList = QList<QRectF>;
            RectangleList rectangles;

            // loop over cells
            for( const auto& cell:cells_ )
            {

                if( ( cell.isEnabled() || showInvalidDaysAction_->isChecked() ) && event.contains( cell.date() ) )
                { rectangles << cell.rect(); }
            }

            // check if rectangles are found
            if( rectangles.empty() ) return pathList;

            // sort and unique list
            std::sort( rectangles.begin(), rectangles.end(), SortRectFTor() );
            rectangles.erase( std::unique( rectangles.begin(), rectangles.end() ), rectangles.end() );

            // unite rectangles to form lines
            RectangleList lines;
            {
                QRectF current;
                for( const auto& rect:rectangles )
                {

                    if( current.isNull() ) current = rect;
                    else if( rect.left() == current.right() && rect.top() == current.top() ) current = current.united( rect );
                    else {
                        lines << current;
                        current = rect;
                    }

                }

                if( !current.isNull() ) lines << current;

            }

            // convert lines to path
            {
                Path current;
                current.setColor( event.color() );
                for( const auto& rect:lines )
                {
                    if( current.rect().isEmpty() ) current.setRect( rect );
                    else if( rect.left() == current.rect().left() && rect.right() == current.rect().right() &&
                        std::abs( current.rect().bottom() - rect.top() ) <= 1 ) current.setRect( current.rect().united( rect ) );
                    else {
                        pathList << current;
                        current.setRect( rect );
                    }
                }

                if( !current.rect().isEmpty() ) pathList << current;

            }

        }

        return pathList;

    }

    //______________________________________________________________________
    void CalendarWidget::_paint( QPaintDevice& device, const QRect& clip )
    {
        Debug::Throw( "CalendarWidget::_paint.\n" );

        QPainter painter( &device );
        painter.setClipRect( clip );
        painter.setRenderHints(QPainter::Antialiasing );
        painter.translate( _margins().topLeft() + QPoint( 0, header_->contentsRect().height() ) );

        // separator
        _paintSeparator( painter );

        // paint events
        if( !_currentEvents().empty() ) { _paintCurrentPathList( painter ); }
        if( !_selectedDays().empty() ) { _paintSelectedPathList( painter ); }

        // text
        _paintText( painter );

        painter.end();

        // loop over children and render them
        #if defined(Q_OS_WIN)
        for( const auto& widget:findChildren<QWidget*>() )
        {
            if( widget->parentWidget() == header_ )
            { widget->render( &device, widget->geometry().topLeft(), widget->rect().intersected( rect().translated( -widget->geometry().topLeft() ) ), 0 ); }
        }
        #endif

        return;

    }

    //_________________________________________
    void CalendarWidget::_paintPathList( QPainter& painter, const Path::List& pathList ) const
    {
        Debug::Throw( "CalendarWidget::_paintPathList.\n" );
        painter.save();

        for( const auto& path:pathList )
        {

            QStyleOptionViewItemV4 option;
            option.initFrom( this );
            option.showDecorationSelected = true;

            const QRectF rectF( path.rect() );
            QRect rect;
            rect.setLeft( std::ceil( rectF.left() ) );
            rect.setRight( std::floor( rectF.right()-0.01 ) );
            rect.setTop( std::ceil( rectF.top() ) );
            rect.setBottom( std::floor( rectF.bottom()-0.01 ) );

            option.rect = rect;
            option.state |= QStyle::State_Selected | QStyle::State_Active;
            option.palette.setCurrentColorGroup( QPalette::Active );

            QColor color( path.color() );
            if( !color.isValid() ) color = option.palette.color( QPalette::Highlight );
            color.setAlpha( 0.5*color.alpha() );
            option.palette.setColor( QPalette::Highlight, color );

            // add alpha channel, needed to make overlaps detectable

            style()->drawPrimitive( QStyle::PE_PanelItemViewItem, &option, &painter, itemView_ );

        }
        painter.restore();

    }

    //___________________________________________________________________________
    void CalendarWidget::_paintSeparator( QPainter& painter ) const
    {
        painter.save();
        painter.setBrush( Qt::NoBrush );
        painter.translate( 0.5, 0.5 );
        painter.setRenderHint(QPainter::Antialiasing, true );

        QColor color( foregroundColor() );
        color.setAlpha( 0.3 * color.alpha() );
        painter.setPen( color );

        if( showWeekNumbersAction_->isChecked() )
        {

            const int left( _cellWidth() );
            const int right( width()-_margins().width() );
            const int top( cellHeight_ );
            const int bottom( height() - _margins().height() - header_->contentsRect().height() );

            QPainterPath path;
            path.moveTo( right, top );
            path.lineTo( left+4, top );
            path.quadTo( left, top, left, top+4 );
            path.lineTo( left, bottom );
            painter.drawPath( path );


        } else {

            const int left( 0 );
            const int right( width()-_margins().width() );
            const int top( cellHeight_ );
            painter.setRenderHint(QPainter::Antialiasing, false );
            painter.drawLine( QPoint( left, top ), QPoint( right, top ) );

        }

        painter.restore();
    }

    //___________________________________________________________________
    void CalendarWidget::_paintText( QPainter& painter ) const
    {

        QPixmap pixmap( size() );
        pixmap.fill( Qt::transparent );
        {
            QPainter painter( &pixmap );
            painter.setPen( foregroundColor() );

            // draw week days name
            for( const auto& cell:weekDayCells_ )
            { painter.drawText( cell.rect(), Qt::AlignCenter, QDate::shortDayName( cell.date().dayOfWeek() ) ); }

            // draw week number
            for( const auto& cell:weekCells_ )
            { painter.drawText( cell.rect(), Qt::AlignCenter, QString::number( cell.date().weekNumber() ) ); }

            // loop over cells
            const QDate today( QDate::currentDate() );
            for( const auto& cell:cells_ )
            {

                if( !cell.isEnabled() ) continue;

                // current date
                if( cell.date() == today )
                {

                    painter.save();
                    painter.setPen( currentDayColor_ );
                    painter.drawText( cell.rect(), Qt::AlignCenter, cell.date().toString( "d" ) );
                    painter.restore();

                } else { painter.drawText( cell.rect(), Qt::AlignCenter, cell.date().toString( "d" ) ); }

            }

            // invalid days
            if( showInvalidDaysAction_->isChecked() )
            {
                painter.setPen( Base::Color( foregroundColor() ).addAlpha( 0.3 ) );
                for( const auto& cell:cells_ )
                {
                    if( cell.isEnabled() ) continue;
                    painter.drawText( cell.rect(), Qt::AlignCenter, cell.date().toString( "d" ) );
                }
            }

        }

        if( shadowOffset() > 0 && shadowColor().isValid() )
        {
            QImage image( pixmap.toImage() );
            Transparency::Effects::shadowBlur(image, shadowOffset(), shadowColor() );
            painter.drawImage(rect().topLeft(), image);
        }

        painter.drawPixmap( rect().topLeft(), pixmap );


    }

    //______________________________________________________________________
    void CalendarWidget::_updateSelection( const QPoint& point, CalendarEvent::SelectionMode mode )
    {

        // find matching date
        QDate date;
        if( ( date = Cell::findCell( cells_, point ).date() ).isValid() ) _updateSelectedDays( date, mode );
        else if( ( date = Cell::findCell( weekCells_, point ).date() ).isValid() )
        {

            for( const auto& cell : Cell::selectWeek( cells_, date ) )
            { _updateSelectedDays( cell.date(), CalendarEvent::SelectionBox ); }

        } else if( ( date = Cell::findCell( weekDayCells_, point ).date() ).isValid() ) {

            for( const auto& cell:Cell::selectWeekDay( cells_, date ) )
            { _updateSelectedDays( cell.date(), CalendarEvent::SelectionBox ); }

        }

        _updateSelectedPathList();
        update();
    }

    //______________________________________________________________________
    void CalendarWidget::_updateSelectedDays( const QDate& currentDate, const CalendarEvent::SelectionMode& mode, bool clear )
    {

        Debug::Throw( "CalendarWidget::_updateSelectedDays.\n" );
        if( clear || !_selectedDate().isValid() )
        {

            selectedDays_.clear();
            _setSelectedDate( currentDate );
            if( currentDate.isValid() )
            {
                CalendarEvent event;
                event.setRange( _selectedDate() );
                event.setSelectionMode( CalendarEvent::SelectionRange );
                event.validate();
                event.generate();
                selectedDays_ << event;
            }

        } else if( currentDate.isValid() ) {

            selectedDays_.clear();

            CalendarEvent event;
            event.setRange( qMin( _selectedDate(), currentDate ), qMax( _selectedDate(), currentDate ) );
            event.setSelectionMode( mode );
            event.validate();
            event.generate();

            selectedDays_ << event;

        }

        return;

    }

    //_________________________________________
    bool CalendarWidget::_setDate( const QDate& date )
    {
        if( date_ == date ) return false;
        date_ = date;
        _updateGeometry();
        _updateCurrentEvents();
        _updateCurrentPathList();
        _updateSelectedPathList();
        header_->label().setText( date_.toString( "MMMM yyyy" ) );
        return true;
    }

    //_________________________________________
    void CalendarWidget::_restartTimer()
    {
        QDateTime now( QDateTime::currentDateTime() );
        int delay = now.secsTo( QDateTime( now.date().addDays( 1 ), QTime( 0, 0, 5 ) ) );
        Debug::Throw() << "CalendarWidget::_restartTimer - delay: " << delay << endl;
        timer_.start( 1000*delay, this );
    }


    //______________________________________________________________________
    bool CalendarWidget::_updateToolTip( const QPoint& position )
    {

        Debug::Throw( "CalendarWidget::_updateToolTip.\n" );

        // clear tooltip widget
        CalendarToolTipWidget* toolTipWidget( static_cast<CalendarToolTipWidget*>( _toolTipWidget() ) );
        toolTipWidget->hide();
        toolTipWidget->clear();

        // get day matching position
        const auto cell( Cell::findCell( cells_, position - _margins().topLeft() - QPoint( 0, header_->contentsRect().height() ) ) );
        const QDate date( cell.date() );
        if( !date.isValid() )
        {
            _toolTipWidget()->hide();
            return false;
        }

        CalendarEvent::List events;
        for( const auto& event:currentEvents_ )
        { if( event.contains( date ) ) events.append( event ); }

        if( !events.empty() )
        {

            for( const auto& event:events )
            { toolTipWidget->addEvent( event ); }

            return true;

        } else {

            return false;

        }

    }

    //______________________________________________________________________
    void CalendarWidget::_checkNotifications()
    {

        Debug::Throw( "CalendarWidget::_checkNotifications.\n" );

        const QDate today( QDate::currentDate() );

        bool modified( false );
        CalendarEvent::List& events( events_.events() );
        for( const auto& event:currentEvents_ )
        {

            // find matching event
            CalendarEvent& persistentEvent( events[events.indexOf( event )] );

            // get range list
            for( const auto& range:event.ranges() )
            {

                // discard events that do not require notification
                if( !event.hasFlag( CalendarEvent::Notify ) ) continue;

                Debug::Throw()
                    << "CalendarWidget::_checkNotifications -"
                    << " flags: " << event.flags()
                    << " distance: " << today.daysTo( range.first )
                    << " checking notification for event " << event.description()
                    << endl;

                // disable notified flags 2 days before
                if( today.daysTo( range.first ) == 2 )
                {

                    persistentEvent.setFlag( CalendarEvent::Notified, false );
                    modified = true;

                } else if( today.daysTo( range.first ) <= 1 && !persistentEvent.hasFlag( CalendarEvent::Notified ) ) {

                    Debug::Throw(0) << "CalendarWidget::_checkNotifications - sending notification for event " << event.description() << endl;

                    // title
                    QString subject( "Calendar Event Automatic Notification" );

                    // generate email body
                    const QString body( event.toString() );

                    // send email
                    CalendarEventSmtpClient* client = new CalendarEventSmtpClient( this, event );
                    connect( client, SIGNAL(sent(CalendarEvent)), SLOT(_eventNotificationSent(CalendarEvent)) );
                    client->setHost( XmlOptions::get().raw( "SMTP_HOST" ), XmlOptions::get().get<int>( "SMTP_PORT" ) );
                    client->send( XmlOptions::get().raw( "SMTP_SENDER" ), XmlOptions::get().raw( "SMTP_RECIPIENT" ), subject, body );

                }

            }

        }

        if( modified )
        {
            events_.write();
            _updateCurrentEvents();
        }

    }

}
