/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "EditAlarmDialog.h"

#include "AlarmDialog.h"
#include "Debug.h"
#include "EditDaysDialog.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "QtUtil.h"
#include "TextEditor.h"
#include "XmlOptions.h"

#include <QLayout>
#include <QLabel>
#include <QPushButton>

namespace Clock
{

    //_______________________________________________________
    EditAlarmDialog::EditAlarmDialog( QWidget* parent ):
        CustomDialog( parent )
    {

        Debug::Throw( "EditAlarmDialog::EditAlarmDialog.\n" );
        setWindowTitle( tr( "Alarm Configuration - Clock" ) );
        setOptionName( "EDIT_ALARM_DIALOG" );

        QVBoxLayout *vLayout( new QVBoxLayout );
        vLayout->setMargin(0);
        vLayout->setSpacing(5);

        QHBoxLayout* hLayout( new QHBoxLayout );
        hLayout->setMargin(0);
        hLayout->setSpacing(10);
        mainLayout().addLayout( hLayout );
        QLabel* label = new QLabel( this );
        label->setMargin(10);
        label->setPixmap( IconEngine::get( IconNames::AlarmLarge ).pixmap( iconSize() ) );
        hLayout->addWidget( label, 0, Qt::AlignTop );
        hLayout->addLayout( vLayout );

        // horizontal layout for hours and minutes
        vLayout->addWidget( enabledCheckbox_ = new QCheckBox( tr( "Enable alarm" ), this ), 1 );
        enabledCheckbox_->setChecked( true );

        QGridLayout* gridLayout( new QGridLayout );
        gridLayout->setMargin(0);
        gridLayout->setSpacing(5);
        gridLayout->setVerticalSpacing(0);
        vLayout->addLayout( gridLayout );

        gridLayout->addWidget( label = new QLabel( tr( "Alarm time:" ), this ), 0, 0, 1, 1 );
        gridLayout->addWidget( timeEditor_ = new QTimeEdit( this ), 0, 1, 1, 1 );
        timeEditor_->setDisplayFormat( "hh:mm" );
        label->setAlignment( Qt::AlignRight|Qt::AlignVCenter );

        gridLayout->addWidget( label = new QLabel( tr( "Days:" ), this ), 1, 0, 1, 1 );
        gridLayout->addWidget( daysLabel_ = new QLabel( this ), 1, 1, 1, 2 );
        gridLayout->addWidget( editDays_ = new QToolButton( this ), 1, 3, 1, 1 );
        editDays_->setIcon( IconEngine::get( IconNames::Edit ) );
        label->setAlignment( Qt::AlignRight|Qt::AlignVCenter );

        connect( editDays_, SIGNAL(pressed()), SLOT(_editDays()) );

        // text
        vLayout->addWidget( new QLabel( tr( "Alarm text:" ), this ) );
        vLayout->addWidget( editor_ = new TextEditor( this ) );
        connect( enabledCheckbox_, SIGNAL(toggled(bool)), timeEditor_, SLOT(setEnabled(bool)) );
        connect( enabledCheckbox_, SIGNAL(toggled(bool)), editor_, SLOT(setEnabled(bool)) );

        // add show button
        QPushButton* button = new QPushButton( IconEngine::get( IconNames::Find ), tr( "Preview" ), this );
        buttonLayout().insertWidget( 1, button );

        connect( button, SIGNAL(clicked()), SLOT(_preview()) );

        adjustSize();
        return;

    }

    //_________________________________________________
    void EditAlarmDialog::setAlarm( const Alarm& alarm )
    {
        Debug::Throw( "EditAlarmDialog::setAlarm.\n" );
        timeEditor_->setTime( QTime( alarm.time().hours(), alarm.time().minutes() ) );

        days_ = alarm.days();
        updateDaysLabel();

        editor_->setText( alarm.text() );
        enabledCheckbox_->setChecked( alarm.enabled() );
        return;
    }

    //___________________________________________________
    Alarm EditAlarmDialog::alarm() const
    {

        Debug::Throw( "EditAlarmDialog::alarm.\n" );
        TimeStamp time( TimeStamp::now() );
        time.setHours( timeEditor_->time().hour() );
        time.setMinutes( timeEditor_->time().minute() );
        time.makeTime();

        // output alarm
        Alarm out( time, editor_->toPlainText(), enabledCheckbox_->isChecked() );
        out.setDays( days_ );

        return out;
    }

    //_________________________________________________
    void EditAlarmDialog::_editDays()
    {
        EditDaysDialog dialog( this );
        dialog.setDays( days_ );
        if( dialog.exec() == QDialog::Rejected ) return;
        days_ = dialog.days();

        updateDaysLabel();
    }

    //____________________________________________________
    void EditAlarmDialog::_preview()
    {
        Debug::Throw( "EditAlarmDialog::_preview.\n" );

        // create and raise alarm dialog
        AlarmDialog* dialog( new AlarmDialog( this, alarm() ) );
        connect( dialog, SIGNAL(repeat(Alarm)), SLOT(_setRepeat(Alarm)) );
        connect( dialog, SIGNAL(closed(Alarm)), SLOT(_unsetRepeat(Alarm)) );
        QtUtil::centerOnDesktop( dialog )->show();

    }

}
