/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "DiskWidget.h"

#include "BaseContextMenu.h"
#include "Color.h"
#include "CppUtil.h"
#include "DiskConfigurationDialog.h"
#include "GridLayout.h"
#include "IconSize.h"
#include "InformationDialog.h"
#include "ToolTipWidget.h"
#include "Util.h"
#include "XmlOptions.h"

#include <QApplication>
#include <algorithm>

namespace Disk
{
    //__________________________________________________________
    DiskWidget::DiskWidget( QWidget *parent ):
        HistogramWidget( parent ),
        diskMonitor_( this )
    {

        Debug::Throw( "Disk::DiskWidget::DiskWidget.\n" );
        setWindowTitle( Util::windowTitle( "Disk Monitor" ) );

        _installActions();
        _configurationAction().setText( tr( "Disk Monitor Settings..." ) );
        _updateConfiguration();

    }

    //______________________________________________________________________
    void DiskWidget::_installContextMenuActions( BaseContextMenu* menu )
    {
        Debug::Throw( "Disk::DiskWidget::_installContextMenuActions.\n" );

        // create parent menu
        HistogramWidget::_installContextMenuActions( menu );

        // actions
        menu->insertSeparator( &_menuSeparator() );
        for( const auto& action:displayActionGroup_->actions() )
        { menu->insertAction( &_menuSeparator(), action ); }

        menu->insertMenu( &_menuSeparator(), &diskMonitor_.menu() );

    }

    //__________________________________________________________
    BaseToolTipWidget* DiskWidget::_createToolTipWidget()
    {
        ToolTipWidget* widget = new ToolTipWidget( this );
        widget->setPixmap( windowIcon().pixmap( IconSize::get( IconSize::Maximum ) ) );

        GridLayout* gridLayout = new GridLayout;
        gridLayout->setMaxCount( 2 );
        gridLayout->setColumnAlignment( 0, Qt::AlignVCenter|Qt::AlignRight );
        gridLayout->setColumnAlignment( 1, Qt::AlignVCenter|Qt::AlignLeft );
        gridLayout->setMargin( 0 );
        gridLayout->setSpacing( 5 );
        widget->layout().addLayout( gridLayout );

        // insert layout items
        GridLayoutItem* item;
        ( item = new GridLayoutItem( widget, gridLayout ) )->setKey( tr( "Host:" ) ); toolTipItems_.append( item );
        ( item = new GridLayoutItem( widget, gridLayout ) )->setKey( tr( "Device:" ) ); toolTipItems_.append( item );
        ( item = new GridLayoutItem( widget, gridLayout ) )->setKey( tr( "Reading:" ) ); toolTipItems_.append( item );
        ( item = new GridLayoutItem( widget, gridLayout ) )->setKey( tr( "Writing:" ) ); toolTipItems_.append( item );

        return widget;
    }

    //______________________________________________________________________
    bool DiskWidget::_updateToolTip()
    {

        Debug::Throw( "Disk::DiskWidget::_updateToolTip.\n" );

        // title
        static_cast<ToolTipWidget*>( _toolTipWidget() )->setTitle( windowTitle() );

        // header
        toolTipItems_[0]->setText( _hostName() );
        toolTipItems_[1]->setText( diskMonitor_.device() );

        // data
        DataSamples::DataList data( diskMonitor_.data() );
        if( !data.empty() )
        {

            toolTipItems_[2]->show(); toolTipItems_[2]->setText( DiskLoad::toString( double(data.back().read())/data.back().time() ) );
            toolTipItems_[3]->show(); toolTipItems_[3]->setText( DiskLoad::toString( double(data.back().write())/data.back().time() ) );

        } else {

            toolTipItems_[2]->hide();
            toolTipItems_[3]->hide();

        }

        return _toolTipWidget()->isHidden();

    }

    //______________________________________________________________________
    void DiskWidget::timerEvent( QTimerEvent* event )
    {

        if( event->timerId() == _updateTimer().timerId() )
        {

            diskMonitor_.update();
            update();

        } else HistogramWidget::timerEvent( event );

        return;
    }

    //______________________________________________________________________
    void DiskWidget::_paint( QPaintDevice& device, const QRect& clip )
    {

        // check margins against window size
        if( !_margins().adjustedRect( rect() ).isValid() )
        {
            InformationDialog( this, tr( "Window size is too small with respect to margins.\n Margins set to 0." ) ).exec();
            _setMargins(0);
        }

        // store size
        QRect rect( _margins().adjustedRect( this->rect() ) );
        const int width( rect.width() );
        const int height( rect.height() );

        // create painter
        QPainter painter( &device );
        painter.setClipRect( clip );
        painter.setRenderHints(QPainter::Antialiasing);

        // retrieve data
        DataSamples::DataList dataList( diskMonitor_.data() );
        while( dataList.size() > width ) dataList.removeFirst();

        if( !dataList.empty() && displayMask_ )
        {

            const bool showWrite( displayMask_ & Data::Write );
            const bool showRead( displayMask_ & Data::Read );

            // running points definition
            double xMin( _margins().left() + 1 );
            double x( xMin );
            double yMin( height + _margins().bottom() + 1 );
            double yMax( yMin );

            QPolygonF writePolygon;
            QPolygonF readPolygon;

            // autoscale painting
            double scale( _getScale( dataList ) );
            for( const auto& data:dataList )
            {

                yMax = yMin;

                if( showWrite )
                {
                    yMax -= qMax( 1.0, scale*data.write()/data.time() );
                    writePolygon << QPointF( x, yMax );
                }

                if( showRead )
                {
                    yMax -= qMax( 1.0, scale*data.read()/data.time() );
                    readPolygon << QPointF( x, yMax );
                }

                x+=1;
            }

            // close polygons
            /*
            polygons are manually 'subtracted' because it should work quicker than
            the general case Qt implementation
            */
            if( showRead )
            {
                if( showWrite ) {
                    for( int i = writePolygon.size()-1; i >= 0; i-- )
                    { readPolygon << writePolygon[i]; }
                } else {
                    readPolygon << QPointF( x-1, yMin );
                    readPolygon << QPointF( xMin, yMin );
                }

                readPolygon << readPolygon.front();
            }

            if( showWrite )
            {
                writePolygon << QPointF( x-1, yMin );
                writePolygon << QPointF( xMin, yMin );
                writePolygon << writePolygon.front();
            }

            // prepare pen
            QPen pen;
            pen.setJoinStyle( Qt::RoundJoin );
            pen.setWidthF( 1.5 );

            // draw read polygon
            if( showRead )
            {
                pen.setColor( colorMap_[Data::Read].foreground() );
                painter.setPen( pen );
                painter.setBrush( colorMap_[Data::Read].background() );
                painter.drawPolygon( readPolygon );

            }

            // draw write polygon
            if( showWrite )
            {
                pen.setColor( colorMap_[Data::Write].foreground() );
                painter.setPen( pen );
                painter.setBrush( colorMap_[Data::Write].background() );
                painter.drawPolygon( writePolygon );
            }

        }

        // draw load
        if( _showLoadAction().isChecked() || dataList.empty() )
        { _paintLoad( painter, rect ); }

        // draw host
        _paintHost( painter, rect );

        painter.end();

    }

    //______________________________________________________________________
    void DiskWidget::resizeEvent( QResizeEvent* event )
    {
        int width( qMax<int>( 0, event->size().width() - _margins().width() ) );
        diskMonitor_.setDataSize( width );
        colorMap_[Data::Read].updateGradient( event->size() );
        colorMap_[Data::Write].updateGradient( event->size() );
        return HistogramWidget::resizeEvent( event );
    }

    //________________________________________________
    void DiskWidget::_configuration()
    {

        Debug::Throw( "Disk::DiskWidget::_configuration.\n" );

        DiskConfigurationDialog dialog;
        _configureFromDialog( dialog );

    }

    //__________________________________________________________
    void DiskWidget::_updateConfiguration()
    {
        Debug::Throw( "Disk::DiskWidget::_updateConfiguration.\n" );

        // colors
        _setDefaultColors();
        if( !XmlOptions::get().get<bool>( "USE_DEFAULT_COLORS" ) )
        {
            colorMap_[Data::Read].update( XmlOptions::get().get<Base::Color>( "READ_COLOR" ) );
            colorMap_[Data::Write].update( XmlOptions::get().get<Base::Color>( "WRITE_COLOR" ) );
        }

        // update mask
        displayMask_ = 0;
        if( XmlOptions::get().get<bool>("SHOW_READ") ) displayMask_ |= Data::Read;
        if( XmlOptions::get().get<bool>("SHOW_WRITE") ) displayMask_ |= Data::Write;
        _updateDisplayActions();

        // base class update
        HistogramWidget::_updateConfiguration();
        update();
        Debug::Throw( "Disk::DiskWidget::_updateConfiguration - done.\n" );

    }

    //____________________________________________________________________
    void DiskWidget::_toggleInverseColors( bool value )
    {

        Debug::Throw( "DiskWidget::_toggleInverseColors" );
        MainWindow::_toggleInverseColors( value );

        // update internal colors
        if( XmlOptions::get().get<bool>( "USE_DEFAULT_COLORS" ) )
        { _setDefaultColors(); }

        update();
    }

    //________________________________________________
    void DiskWidget::_toggleDisplay( QAction* action )
    {

        Debug::Throw( "Disk::DiskWidget::_toggleDisplay.\n" );
        Data::Type bit( displayActions_[action] );
        if( action->isChecked() ) displayMask_ |= bit;
        else displayMask_ &= ~bit;
        XmlOptions::get().set<bool>( "SHOW_READ", displayMask_ & Data::Read );
        XmlOptions::get().set<bool>( "SHOW_WRITE", displayMask_ & Data::Write );

        update();

    }

    //______________________________________________________________________
    void DiskWidget::_installActions()
    {

        Debug::Throw( "Disk::DiskWidget::_installActions.\n" );

        // display
        displayActionGroup_ = new QActionGroup( this );
        displayActionGroup_->setExclusive( false );
        displayActions_ = Base::makeT<DisplayActionMap>( {
            { displayActionGroup_->addAction( tr( "Read" ) ), Data::Read },
            { displayActionGroup_->addAction( tr( "Write" ) ), Data::Write }
        });

        for( auto&& iter = displayActions_.begin(); iter != displayActions_.end(); ++iter )
        { iter.key()->setCheckable( true ); }

        _updateDisplayActions();
        connect( displayActionGroup_, SIGNAL(triggered(QAction*)), SLOT(_toggleDisplay(QAction*)) );

    }

    //______________________________________________________________________
    void DiskWidget::_updateDisplayActions()
    {
        Debug::Throw( "Disk::DiskWidget::_updateDisplayActions.\n" );
        for( auto&& iter = displayActions_.begin(); iter != displayActions_.end(); ++iter )
        {
            iter.key()->setChecked( displayMask_ & iter.value() );
            iter.key()->setIcon( colorMap_[iter.value()].icon() );
        }
    }

    //_________________________________________________________
    double DiskWidget::_getScale( const DataSamples::DataList& data ) const
    {
        Debug::Throw( "Disk::DiskWidget::_getScale.\n" );

        Q_ASSERT( !data.empty() );

        // get maximum
        auto iter( std::max_element( data.begin(), data.end(), DiskLoad::LessFTor( displayMask_ ) ) );
        double maximum = double( iter->sum( displayMask_ ) )/iter->time();

        int height( qMax<int>( 1, DiskWidget::height() - _margins().height() ) );
        return (maximum) ? double( height )/maximum:1;
    }

    //_________________________________________________________
    void DiskWidget::_paintLoad( QPainter& painter, QRect& rect )
    {
        Debug::Throw( "Disk::DiskWidget::_paintLoad.\n" );

        // prepare load text
        QString buffer;
        DataSamples::DataList data( diskMonitor_.data() );
        if( !data.empty() ) {

            buffer = QString( tr( "Writing: %1" ) ).arg( DiskLoad::toString( double(data.back().write())/data.back().time() ) );
            _paintText( painter, buffer, rect );

            buffer = QString( tr( "Reading: %1" ) ).arg( DiskLoad::toString( double(data.back().read())/data.back().time() ) );
            _paintText( painter, buffer, rect );

        } else {

            buffer = QString( tr( "Acquiring (%1)" ) ).arg( diskMonitor_.device()  );
            _paintText( painter, buffer, rect );

        }

        return;

    }

    //_________________________________________________________
    void DiskWidget::_setDefaultColors()
    {
        Debug::Throw() << "CpuWidget::_setDefaultColor" << endl;
        colorMap_[Data::Read].update( Base::Color( foregroundColor() ).merge( shadowColor(), 0.5 ) );
        colorMap_[Data::Write].update( Base::Color( foregroundColor() ).merge( shadowColor(), 0.75 ) );
    }

}
