/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Anchor.h"
#include "ApplicationDataDialog.h"
#include "Debug.h"
#include "GeometryApplication.h"
#include "GeometryCalculationObject.h"
#include "GeometryConfirmationDialog.h"
#include "GeometryDialog.h"
#include "GridLayout.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "InformationDialog.h"
#include "Singleton.h"
#include "TreeView.h"
#include "XmlOptions.h"

#include <QApplication>
#include <QDesktopWidget>
#include <QGroupBox>
#include <QLabel>

#include <algorithm>

//___________________________________________________
GeometryDialog::GeometryDialog( QWidget* parent ):
CustomDialog( parent, OkButton|CancelButton ),
OptionWidgetList( this )
{

    Debug::Throw( "GeometryDialog::GeometryDialog.\n" );
    setOptionName( "GEOMETRY_DIALOG" );

    QGroupBox* group;
    mainLayout().addWidget( group = new QGroupBox( tr( "Applications" ), this ) );

    QHBoxLayout* hLayout( new QHBoxLayout );
    hLayout->setMargin(10);
    hLayout->setSpacing(5);
    group->setLayout( hLayout );

    // list
    hLayout->addWidget( list_ = new TreeView( group ) );
    list_->setModel( &model_ );
    list_->setSortingEnabled( false );
    list_->setOptionName( "APPLIST" );

    list_->setDragEnabled(true);
    list_->setAcceptDrops(true);

    connect( &model_, SIGNAL(itemOrderChanged(int,int)), SLOT(_reorder(int,int)) );

    // buttons
    QVBoxLayout* vLayout( new QVBoxLayout );
    vLayout->setMargin(0);
    hLayout->addLayout( vLayout );

    QPushButton* button;
    vLayout->addWidget( button = new QPushButton( tr( "Sort" ), group ) );
    connect( button, SIGNAL(clicked()), SLOT(_sort()) );
    button->setIcon( IconEngine::get( IconNames::Sort ) );
    button->setToolTip( tr( "Sort applications based on alignment and current geometry" ) );

    vLayout->addWidget( edit_ = new QPushButton( tr( "Edit" ), group ) );
    connect( edit_, SIGNAL(clicked()), SLOT(_edit()) );
    edit_->setIcon( IconEngine::get( IconNames::Edit ) );
    edit_->setToolTip( tr( "Edit selected application geometry" ) );

    vLayout->addWidget( up_ = new QPushButton( tr( "Move Up" ), group ) );
    connect( up_, SIGNAL(clicked()), SLOT(_up()) );
    up_->setIcon( IconEngine::get( IconNames::Up ) );
    up_->setToolTip( tr( "Move up selected application" ) );

    vLayout->addWidget( down_ = new QPushButton( tr( "Move Down" ), group ) );
    connect( down_, SIGNAL(clicked()), SLOT(_down()) );
    down_->setIcon( IconEngine::get( IconNames::Down ) );
    down_->setToolTip( tr( "Move down selected application" ) );

    vLayout->addWidget( reload_ = new QPushButton( tr( "Reload" ), group ) );
    connect( reload_, SIGNAL(clicked()), SLOT(reload()) );
    reload_->setIcon( IconEngine::get( IconNames::Reload ) );
    reload_->setToolTip( tr( "Reload applications geometry" ) );

    QFrame* frame;
    vLayout->addWidget( frame = new QFrame( group ) );
    frame->setFrameStyle( QFrame::HLine );

    QPushButton* configurationButton;
    vLayout->addWidget( configurationButton = new QPushButton( tr( "Configure" ), group ) );
    connect(configurationButton, SIGNAL(clicked()), &static_cast<GeometryApplication*>(Base::Singleton::get().application())->configurationAction(), SLOT(trigger()));
    configurationButton->setIcon( IconEngine::get( IconNames::Configure ) );
    configurationButton->setToolTip( tr( "Configure Transparency-settings" ) );

    vLayout->addStretch(1);

    // set connections
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateButtons()) );
    connect( list_->selectionModel(), SIGNAL(currentChanged(QModelIndex,QModelIndex)), SLOT(_updateButtons()) );
    connect( list_, SIGNAL(activated(QModelIndex)), SLOT(_edit()) );

    mainLayout().addWidget( group = new QGroupBox( tr( "Automatic Placement" ), this ) );

    GridLayout* gridLayout = new GridLayout;
    gridLayout->setMargin(10);
    gridLayout->setSpacing(5);
    gridLayout->setMaxCount(2);
    group->setLayout( gridLayout );

    gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );

    QLabel* label;
    gridLayout->addWidget( label = new QLabel( tr( "Spacing:" ), this ) );
    gridLayout->addWidget( spacingSpinBox_ = new OptionSpinBox( this, "SETTINGS_SPACING", false ) );
    spacingSpinBox_->setSuffix( tr( "px" ) );
    spacingSpinBox_->setMinimum( -64 );
    spacingSpinBox_->setMaximum( 64 );
    label->setBuddy( spacingSpinBox_ );
    addOptionWidget( spacingSpinBox_ );

    gridLayout->addWidget( label = new QLabel( tr( "Horizontal offset:" ), this ) );
    gridLayout->addWidget( horizontalOffsetSpinBox_ = new OptionSpinBox( this, "SETTINGS_OFFSET_HORIZONTAL", false ) );
    horizontalOffsetSpinBox_->setSuffix( tr( "px" ) );
    horizontalOffsetSpinBox_->setMinimum( 0 );
    horizontalOffsetSpinBox_->setMaximum( 64 );
    label->setBuddy( horizontalOffsetSpinBox_ );
    addOptionWidget( horizontalOffsetSpinBox_ );

    gridLayout->addWidget( label = new QLabel( tr( "Vertical offset:" ), this ) );
    gridLayout->addWidget( verticalOffsetSpinBox_ = new OptionSpinBox( this, "SETTINGS_OFFSET_VERTICAL", false ) );
    verticalOffsetSpinBox_->setSuffix( tr( "px" ) );
    verticalOffsetSpinBox_->setMinimum( 0 );
    verticalOffsetSpinBox_->setMaximum( 64 );
    label->setBuddy( verticalOffsetSpinBox_ );
    addOptionWidget( verticalOffsetSpinBox_ );

    gridLayout->addWidget( label = new QLabel( tr( "Orientation:" ), this ) );
    gridLayout->addWidget( alignmentComboBox_ = new OptionComboBox( this, "SETTINGS_ALIGNMENT" ) );
    alignmentComboBox_->setUseValue( false );
    alignmentComboBox_->addItems( QStringList()
        << tr( "Horizontal" )
        << tr( "Vertical" ) );
    label->setBuddy( alignmentComboBox_ );
    addOptionWidget( alignmentComboBox_ );

    anchorWidget_ = new OptionAnchorWidget( this, "SETTINGS_ANCHOR", gridLayout );
    addOptionWidget( anchorWidget_ );
    gridLayout->setColumnStretch( 1, 1 );

    // apply button
    apply_ = buttonBox().addButton( QDialogButtonBox::Apply );
    #if defined( Q_OS_WIN )
    apply_->setIcon( IconEngine::get( IconNames::DialogAccept ) );
    #endif
    connect( apply_, SIGNAL(clicked()), SLOT(_apply()) );

    // update buttons
    _updateButtons();

    connect( alignmentComboBox_, SIGNAL(currentIndexChanged(int)), SLOT(_sort(int)) );

    connect( &okButton(), SIGNAL(clicked()), SLOT(_applyUnchecked()) );
    connect( &cancelButton(), SIGNAL(clicked()), SLOT(_restoreDefaults()) );

    connect( Base::Singleton::get().application(), SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( qApp, SIGNAL(aboutToQuit()), SLOT(_saveConfiguration()) );

}

//____________________________________________________
void GeometryDialog::add( ApplicationData data )
{
    model_.add( data );
    _sort();
    list_->resizeColumns();
}

//____________________________________________________
void GeometryDialog::reload()
{
    Debug::Throw( "GeometryDialog::reload.\n" );
    model_.clear();
    backup_.clear();
    emit geometryRequested();

}

//______________________________________________________________________
void GeometryDialog::_updateButtons()
{
    Debug::Throw( "GeometryDialog::_updateButtons.\n" );
    ApplicationDataModel::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    bool hasSelection( !selection.empty() );

    edit_->setEnabled( hasSelection );
    up_->setEnabled( hasSelection && !list_->selectionModel()->isSelected( model_.index( 0, 0 ) ) );
    down_->setEnabled( hasSelection && !list_->selectionModel()->isSelected( model_.index( model_.rowCount()-1, 0 ) ) );
}

//____________________________________________________
void GeometryDialog::_edit()
{

    Debug::Throw( "GeometryDialog::_edit.\n" );

    // retrieve selection
    QModelIndex current( list_->selectionModel()->currentIndex() );
    Q_ASSERT( current.isValid() );

    ApplicationData currentData( model_.get( current ) );
    ApplicationDataDialog dialog( this, currentData );
    if( !dialog.centerOnWidget( parentWidget() ).exec() ) return;

    ApplicationData data( dialog.data() );
    if( data.geometry() == currentData.geometry() ) return;

    // propagate order
    emit geometryChanged( data );

    return;

}

//______________________________________________________________________
void GeometryDialog::_sort( int orientation )
{
    Debug::Throw() << "GeometryDialog::_sort - orientation: " << orientation << endl;
    ApplicationData::List applications( model_.get() );
    ApplicationData::sort( applications, static_cast<ApplicationData::Orientation>(orientation) );
    model_.set( applications );
}

//______________________________________________________________________
void GeometryDialog::_sort()
{ _sort( alignmentComboBox_->currentIndex() ); }

//_________________________________________________________
void GeometryDialog::_up()
{

    Debug::Throw( "GeometryDialog::_up.\n" );
    ApplicationDataModel::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() ) return;

    // retrieve selected indexes in list and store in model
    QModelIndexList selectedIndexes( list_->selectionModel()->selectedRows() );
    ApplicationDataModel::List selectedData( model_.get( selectedIndexes ) );

    ApplicationDataModel::List currentData( model_.get() );
    ApplicationDataModel::List newData;

    for( const auto& applicationData:currentData )
    {

        // check if new list is not empty, current index is selected and last index is not.
        // if yes, move.
        if(
            !( newData.empty() ||
            selectedIndexes.indexOf( model_.index( applicationData ) ) == -1 ||
            selectedIndexes.indexOf( model_.index( newData.back() ) ) != -1
            ) )
        {
            ApplicationData last( newData.back() );
            newData.removeLast();
            newData << applicationData;
            newData << last;
        } else newData << applicationData;

    }

    model_.set( newData );

    // restore selection
    list_->selectionModel()->select( model_.index( selectedData.front() ),  QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
    for( const auto& applicationData:selectedData )
    { list_->selectionModel()->select( model_.index( applicationData ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

    return;

}

//_________________________________________________________
void GeometryDialog::_down()
{

    Debug::Throw( "GeometryDialog::_down.\n" );
    ApplicationDataModel::List selection( model_.get( list_->selectionModel()->selectedRows() ) );
    if( selection.empty() ) return;

    // retrieve selected indexes in list and store in model
    QModelIndexList selectedIndexes( list_->selectionModel()->selectedIndexes() );
    ApplicationDataModel::List selectedData( model_.get( selectedIndexes ) );

    ApplicationDataModel::List newData;
    ApplicationDataModel::ListIterator iter( model_.get() );
    iter.toBack();
    while( iter.hasPrevious() )
    {
        const ApplicationData& current( iter.previous() );

        // check if new list is not empty, current index is selected and last index is not.
        // if yes, move.
        if(
            !( newData.empty() ||
            selectedIndexes.indexOf( model_.index( current ) ) == -1 ||
            selectedIndexes.indexOf( model_.index( newData.front() ) ) != -1
            ) )
        {

            ApplicationData first( newData.front() );
            newData.removeFirst();
            newData.prepend( current );
            newData.prepend( first );

        } else newData.prepend( current );
    }

    model_.set( newData );

    // restore selection
    list_->selectionModel()->select( model_.index( selectedData.front() ),  QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
    for( const auto& applicationData:selectedData )
    { list_->selectionModel()->select( model_.index( applicationData ), QItemSelectionModel::Select|QItemSelectionModel::Rows ); }

    return;

}

//_________________________________________________________
void GeometryDialog::_reorder( int oldRow, int newRow )
{

    Debug::Throw() << "GeometryDialog::_reorder - old: " << oldRow << " new: " << newRow << endl;
    if( newRow > oldRow ) newRow--;

    ApplicationDataModel::List dataList( model_.get() );
    dataList.move( oldRow, newRow );
    model_.set( dataList );

    // restore selection
    list_->selectionModel()->select( model_.index( newRow, 0 ),  QItemSelectionModel::Clear|QItemSelectionModel::Select|QItemSelectionModel::Rows );
    return;

}

//____________________________________________________
void GeometryDialog::_apply( bool warn )
{

    Debug::Throw( "GeometryDialog::_apply.\n" );

    // get application from model and save backup
    ApplicationData::List applications( model_.get() );
    if( backup_.empty() ) backup_ = applications;

    // write options to xml
    OptionWidgetList::write( XmlOptions::get() );

    // create calculation object
    GeometryCalculationObject calculationObject;
    calculationObject.setApplications( applications );

    auto desktop( qApp->desktop() );
    calculationObject.setDesktopGeometry( desktop->availableGeometry( desktop->screenNumber( this ) ) );

    calculationObject.calculate();

    // get new applications
    applications = calculationObject.applications();

    // get model current set
    ApplicationDataModel::List current( model_.get() );
    ApplicationDataModel::List newApplications;

    // update model
    model_.set( applications );

    // send new geometries to client
    for( const auto& data:applications )
    {

        auto currentIter( std::find( current.begin(), current.end(), data ) );
        if( currentIter != current.end() && (currentIter->geometry() != data.geometry() || currentIter->order() != data.order() ) )
        { newApplications.append( data ); }

    }

    // check, ask confirmation and emit signals
    if( newApplications.empty() )
    {
        if( warn )
        { InformationDialog( this, tr( "There are no changes to apply with respect to current geometry. <Apply> Canceled." ) ).exec(); }
        return;
    }

    if( !GeometryConfirmationDialog( this, newApplications ).exec() ) return;

    for( const auto& data:newApplications )
    { emit geometryChanged( data ); }

}

//____________________________________________________
void GeometryDialog::_restoreDefaults()
{

    if( backup_.empty() ) return;
    ApplicationDataModel::List current( model_.get() );
    for( const auto& applicationData:backup_ )
    {

        auto currentIter( std::find( current.begin(), current.end(), applicationData ) );
        if( currentIter != current.end() && currentIter->geometry() != applicationData.geometry() )
        { emit geometryChanged( applicationData ); }

    }

}

//____________________________________________________
void GeometryDialog::_updateConfiguration()
{ OptionWidgetList::read( XmlOptions::get() ); }

//____________________________________________________
void GeometryDialog::_saveConfiguration()
{ OptionWidgetList::write( XmlOptions::get() ); }
