/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "PictureApplication.h"

#include "ImageFileDialog.h"
#include "InformationDialog.h"
#include "PictureListDialog.h"
#include "PictureWidget.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "PictureParser.h"
#include "QtUtil.h"
#include "Util.h"
#include "XmlOptions.h"
#include "Config.h"

#include <QApplication>

//____________________________________________
PictureApplication::PictureApplication( CommandLineArguments arguments ) :
    BaseApplication( nullptr, arguments ),
    Counter( "PictureApplication" )
{
    Debug::Throw( "PictureApplication::PictureApplication.\n" );
    qApp->setQuitOnLastWindowClosed( false );
}

//____________________________________________
bool PictureApplication::realizeWidget()
{
    Debug::Throw( "PictureApplication::realizeWidget.\n" );

    // check if the method has already been called.
    if( !BaseApplication::realizeWidget() ) return false;

    // actions
    _installActions();

    // read images from arguments
    QStringList filenames( CommandLineParser().parse( _arguments() ).orphans() );

    // if no images are provided, use the ones stored in resource file
    if( filenames.isEmpty() ) _readPictures();
    else {

        for( const auto& filename:filenames )
        {

            File file( filename );
            file.expand();

            // check file existence
            if( !file.exists() ) continue;
            Debug::Throw(0) << "PictureApplication::realizeWidget - loading: " << file << endl;

            Pictures::PictureAttributes attributes;
            attributes.setFilename( file );
            Pictures::PictureWidget* widget = new Pictures::PictureWidget( 0, attributes );
            widget->adjustSizeToImage();
            QtUtil::centerOnDesktop( widget );
            _addPicture( widget );
        }

    }

    // configuration
    emit configurationChanged();

    // if no picture is set, creates an empty one
    if( pictureWidgets_.empty() )
    {

        Pictures::PictureWidget* widget( new Pictures::PictureWidget() );
        _addPicture( widget );
        widget->show();

    } else {

        bool hasVisible( false );
        for( auto& pictureWidget:pictureWidgets_ )
        {
            if( pictureWidget->attributes().isVisible() )
            {
                hasVisible = true;
                pictureWidget->show();
            }
        }

        // if no picture is visible, raise the window list dialog instead
        if( !hasVisible )
        {
            InformationDialog( 0, tr( "All registered pictures are presently hidden.\n Select the one you want to show from the list." ) ).centerOnWidget( qApp->activeWindow() ).exec();
            _showPictureList();
        }

    }

    return true;

}

//____________________________________________
void PictureApplication::usage() const
{
    _usage( "transparent-pictures" );
    commandLineParser().usage();
    return;
}

//_______________________________________________
QString PictureApplication::applicationName() const
    { return "Picture Viewer"; }

//_______________________________________________
QString PictureApplication::applicationVersion() const
{ return VERSION; }

//_______________________________________________
void PictureApplication::_configuration()
{}

//_______________________________________________
void PictureApplication::_raiseAll()
{
    Debug::Throw( "PictureApplication::_raiseAll.\n" );

    for( auto& pictureWidget:pictureWidgets_ )
    { if( pictureWidget->attributes().isVisible() ) pictureWidget->uniconify(); }

}

//_______________________________________________
void PictureApplication::_updateConfiguration()
{
    showFileNameAction_->setChecked( XmlOptions::get().get<bool>( "SHOW_FILENAME" ) );
    keepAspectRatioAction_->setChecked( XmlOptions::get().get<bool>( "KEEP_ASPECT_RATIO" ) );
}

//________________________________________________
bool PictureApplication::_processCommand( Server::ServerCommand command )
{

    Debug::Throw( "PictureApplication::_processCommand.\n" );
    if( BaseApplication::_processCommand( command ) ) return true;
    if( command.command() == Server::ServerCommand::Raise )
    {
        for( auto& pictureWidget:pictureWidgets_ )
        { pictureWidget->uniconify(); }

        return true;

    } else return false;

}

//_______________________________________________
void PictureApplication::_writePictures( bool check )
{

    Debug::Throw( "PictureApplication::_writePictures.\n" );

    // if taskBars have not been initialized, do nothing
    if( !_realized() ) return;

    // check taskBar
    if( pictureWidgets_.empty() ) return;

    // try open XML file
    File file( XmlOptions::get().raw( "RC_FILE" ) );

    Pictures::PictureParser parser;
    for( const auto& widget:pictureWidgets_ )
    {
        if( widget->attributes().filename() != Pictures::PictureAttributes::NoPicture )
        { parser.addAttributes( widget->attributes() ); }
    }

    bool written( parser.write( file ) );
    if( (!written) && check )
    {
        const QString buffer = QString( tr( "Unable to write to file '%1'" ) ).arg( file );
        InformationDialog( 0, buffer ).exec();
    }

    return;

}

//_______________________________________________________________________________________________
void PictureApplication::_newPicture()
{
    Debug::Throw( "PictureApplication::_newPicture.\n" );

    // create dialog
    ImageFileDialog dialog(0);
    dialog.setFileMode( QFileDialog::ExistingFile );
    dialog.setAcceptMode( QFileDialog::AcceptOpen );
    QtUtil::centerOnWidget( &dialog, qApp->activeWindow() );

    // try get current picture widget
    Pictures::PictureWidget* current( qobject_cast<Pictures::PictureWidget*>( qApp->activeWindow() ) );
    if( current && current->attributes().filename() != Pictures::PictureAttributes::NoPicture )
    {
        File path( File( current->attributes().filename() ).path() );
        if( path.exists() && path.isDirectory() ) dialog.setDirectory( path );
    }

    if( dialog.exec() == QDialog::Rejected ) return;

    // retrieve selected files
    QStringList files( dialog.selectedFiles() );

    // create attributes
    Pictures::PictureAttributes attributes;

    // copy size from current widget
    if( current && current->attributes().size().isValid() )
    {
        attributes.setSize( current->attributes().size() );
        attributes.setPosition( QtUtil::centerOnDesktop( attributes.size() ) );
    }

    // check file size
    if( files.size() == 1 )
    { attributes.setFilename( File( files.front() ) ); }

    // create new picture and save in list
    Pictures::PictureWidget* widget( new Pictures::PictureWidget( 0, attributes ) );
    _addPicture( widget );
    widget->show();

    return;

}

//_______________________________________________________________________________________________
void PictureApplication::_showPictureList()
{
    Debug::Throw( "PictureApplication::_showPictureList.\n" );

    Pictures::PictureAttributes::List attributes;
    for( const auto& widget:pictureWidgets_ )
    { attributes << widget->attributes(); }

    Pictures::PictureListDialog dialog( 0, attributes );
    if( dialog.centerOnWidget( qApp->activeWindow() ).exec() )
    {

        // retrieve attributes from dialog
        attributes = dialog.attributes();

        // delete widgets that are not found in the new list of attributes
        Pictures::PictureWidget::List tmpWidgets;
        for( const auto& widget:pictureWidgets_ )
        {
            Pictures::PictureAttributes::List::iterator attributesIter( std::find_if( attributes.begin(), attributes.end(), Pictures::PictureWidget::SamePositionFTor( widget->attributes().position() ) ) );
            if( attributesIter == attributes.end() )
            {

                widget->deleteLater();

            } else {

                // update with attributes
                widget->setAttributes( *attributesIter );
                tmpWidgets << widget;

                if( widget->attributes().isVisible() ) widget->show();
                else widget->hide();

                // remove attributes from list
                attributes.erase( std::remove( attributes.begin(), attributes.end(), *attributesIter ), attributes.end() );

            }

        }

        // update widget list
        pictureWidgets_ = tmpWidgets;

        // add remaining attributes
        for( Pictures::PictureAttributes::List::iterator iter = attributes.begin(); iter != attributes.end(); ++iter )
        {
            Pictures::PictureWidget* widget( new Pictures::PictureWidget( 0, *iter ) );
            _addPicture( widget );
            if( widget->attributes().isVisible() ) widget->show();
        }

    }

    // check if at least one widget is visible
    for( const auto& widget:pictureWidgets_ )
    { if( widget->attributes().isVisible() ) return; }

    // quit if all windows are hidden
    qApp->quit();

    return;

}

//_______________________________________________
void PictureApplication::_pictureDestroyed( QObject* object )
{

    Debug::Throw( "PictureApplication::_pictureDestroyed.\n" );

    // remove object from list of widgets
    pictureWidgets_.erase( std::remove( pictureWidgets_.begin(), pictureWidgets_.end(), object ), pictureWidgets_.end() );

    // quit if no pictures remain
    if( pictureWidgets_.empty() ) qApp->quit();

}

//________________________________________________
void PictureApplication::_toggleKeepAspectRatio( bool state )
{

    Debug::Throw( "PictureApplication::_toggleKeepAspectRatio.\n" );
    XmlOptions::get().set<bool>( "KEEP_ASPECT_RATIO", state );

}

//________________________________________________
void PictureApplication::_toggleShowFilename( bool state )
{

    Debug::Throw( "PictureApplication::_toggleShowFilename.\n" );
    XmlOptions::get().set<bool>( "SHOW_FILENAME", state );

}

//_______________________________________________
void PictureApplication::_pictureHidden()
{
    Debug::Throw( "PictureApplication::_pictureHidden.\n" );

    for( const auto& widget:pictureWidgets_ )
    { if( widget->attributes().isVisible() ) return; }

    qApp->quit();

}

//_______________________________________________
void PictureApplication::_addPicture( Pictures::PictureWidget* widget )
{
    Debug::Throw( "PictureApplication::_addPicture.\n" );

    connect( widget, SIGNAL(pictureHidden()), SLOT(_pictureHidden()) );
    connect( widget, SIGNAL(destroyed(QObject*)), SLOT(_pictureDestroyed(QObject*)) );

    widget->addAction( &closeAction() );

    pictureWidgets_ << widget;

}

//_______________________________________________
void PictureApplication::_readPictures()
{
    Debug::Throw( "PictureApplication::_readPictures.\n" );

    // read tasks from db file
    const File file( XmlOptions::get().raw("RC_FILE" ) );
    Pictures::PictureParser parser;
    parser.read( file );

    // check errors
    if( parser.xmlError() )
    {
        QString buffer = QString( tr( "Errors occured while parsing file '%1':\n" ) ).arg( file );
        buffer += parser.xmlError().toString();
        InformationDialog( 0, buffer ).exec();
    }

    // clear picture windows
    for( const auto& widget:pictureWidgets_ ) widget->deleteLater();
    pictureWidgets_.clear();

    const Pictures::PictureAttributes::List& attributesList( parser.attributes() );
    for( const auto& attributes:attributesList )
    { _addPicture( new Pictures::PictureWidget( 0, attributes ) ); }

    return;

}

//_______________________________________________________________________________________________
void PictureApplication::_installActions()
{

    Debug::Throw( "PictureApplication::_installActions.\n" );

    // rename about action
    aboutAction().setText( "About Picture Viewer" );

    // configuration
    connect( this, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( qApp, SIGNAL(aboutToQuit()), SLOT(_writePictures()) );

    // raise all
    raiseAllAction_ = new QAction( IconEngine::get( IconNames::RaiseAll ), tr( "Raise All" ), this );
    raiseAllAction_->setToolTip( tr( "Raise all transparent windows" ) );
    connect( raiseAllAction_, SIGNAL(triggered()), SLOT(_raiseAll()) );

    // new picture action
    newPictureAction_ = new QAction( IconEngine::get( IconNames::Add ), tr( "New" ), this );
    newPictureAction_->setToolTip( tr( "Add new picture" ) );
    newPictureAction_->setShortcut( QKeySequence::New );
    connect( newPictureAction_, SIGNAL(triggered()), SLOT(_newPicture()) );

    // picture list
    pictureListAction_ = new QAction( IconEngine::get( IconNames::Find ), tr( "Picture List" ), this );
    pictureListAction_->setToolTip( tr( "Show list of pictures" ) );
    connect( pictureListAction_, SIGNAL(triggered()), SLOT(_showPictureList()) );

    keepAspectRatioAction_ = new QAction( tr( "Keep Aspect Ratio" ), this );
    keepAspectRatioAction_->setCheckable( true );
    keepAspectRatioAction_->setChecked( false );
    connect( keepAspectRatioAction_, SIGNAL(toggled(bool)), SLOT(_toggleKeepAspectRatio(bool)) );

    showFileNameAction_ = new QAction( tr( "Show Filename" ), this );
    showFileNameAction_->setCheckable( true );
    showFileNameAction_->setChecked( false );
    connect( showFileNameAction_, SIGNAL(toggled(bool)), SLOT(_toggleShowFilename(bool)) );

}
