/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TransparentWindow.h"

#include "Anchor.h"
#include "BaseApplication.h"
#include "BaseContextMenu.h"
#include "Color.h"
#include "CompositeEngine.h"
#include "ConfigurationDialog.h"
#include "CustomPixmap.h"
#include "DebugMenu.h"
#include "DefaultHelpText.h"
#include "HelpManager.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "IconSize.h"
#include "OptionAnchorWidget.h"
#include "QtUtil.h"
#include "RoundedRegion.h"
#include "Singleton.h"
#include "SvgEngine.h"
#include "SvgHelpText.h"
#include "TransparencyHelpText.h"
#include "WinUtil.h"
#include "XmlOptions.h"
#include "XcbUtil.h"
#include "XcbThread.h"

#include <QApplication>
#include <QDesktopWidget>
#include <QPainter>
#include <QTimer>

#if HAVE_XCB
#include <xcb/xcb.h>
#endif

//_________________________________________________________________
TransparentWindow::TransparentWindow( QWidget* parent ):
    Transparency::TransparentWidget( parent, Qt::FramelessWindowHint ),
    menuSeparator_( 0 )
{
    Debug::Throw( "TransparentWindow::TransparentWindow.\n" );

    // actions
    _installActions();
    _updateConfiguration();

    // help system
    help_ = new Base::HelpManager( this );
    help_->install( Transparency::helpText );
    help_->install( Svg::helpText );
    help_->install( Base::helpText, false );

    setProperty( "_kde_no_blur_behind", true );

    connect( &Svg::SvgEngine::get(), SIGNAL(changed()), SLOT(_updateMargins()) );
    connect( &Svg::SvgEngine::get(), SIGNAL(changed()), SLOT(setBackgroundChanged()) );
    connect( qApp->desktop(), SIGNAL(workAreaResized(int)), SLOT(_loadGeometryDelayed()) );

    // xcb watcher thread
    auto thread = new XcbThread( this );
    connect( thread, SIGNAL(stateChanged(WId)), SLOT(_processStateChanged(WId)));
    connect( thread, SIGNAL(desktopChanged(WId,quint32)), SLOT(_processDesktopChanged(WId,quint32)));
    thread->setWindowId( winId() );
    thread->start();

}

//____________________________________________
TransparentWindow::~TransparentWindow()
{

    // terminate all xcb threads
    for( auto&& thread:findChildren<XcbThread*>() )
    { thread->terminate(); }

}

//_____________________________________________________
void TransparentWindow::uniconify()
{ QtUtil::uniconify( this ); }

//________________________________________________
void TransparentWindow::processOption( const XmlOption& option )
{

    Debug::Throw()
        << "TransparentWindow::processOption -"
        << " recieved option: " << option.name()
        << " value: " << option.raw()
        << endl;

    QAction* action(0);
    if( option.name() == "STAYS_ON_TOP" ) action = staysOnTopAction_;
    else if( option.name() == "LOCK_POSITION" ) action = lockPositionAction_;
    else if( option.name() == "HIDE_FROM_TASKBAR" ) action = hideFromTaskBarAction_;
    else if( option.name() == "STICKY" ) action = stickyAction_;
    else if( option.name() == "ROUND_CORNERS" ) action = roundCornersAction_;
    else if( option.name() == "USE_SVG" ) action = useSvgAction_;
    else if( option.name() == "TRANSPARENCY_INVERSE_COLORS" ) action = &inverseColorsAction();
    else Debug::Throw(0) << "TransparentWindow::processOption - option " << option.name() << " not implemented" << endl;

    if( action ) action->setChecked( option.get<bool>() );

}

//________________________________________________
QRect TransparentWindow::innerGeometry() const
{ return _outerPadding().adjustedRect( geometry() ); }

//________________________________________________
void TransparentWindow::setInnerGeometry( const QRect& innerGeometry )
{

    // set geometry, properly accounting for outer padding
    const Base::Margins outerPadding( _outerPadding() );
    const QRect geometry( innerGeometry.adjusted( -outerPadding.left(), -outerPadding.top(), outerPadding.right(), outerPadding.bottom() ) );
    if( geometry != this->geometry() ) setGeometry( geometry );

}

//________________________________________________
void TransparentWindow::setOrder( int order )
{
    if( order != order_ )
    {
        order_ = order;
        XmlOptions::get().set<int>( "ORDER", order_ );
    }
}

//__________________________________________________________
void TransparentWindow::_processStateChanged( WId )
{
    #if HAVE_XCB
    // restore sticky if unset
    if( XcbUtil::get().isSupported( XcbDefines::_NET_WM_STATE_STICKY )
        && stickyAction_->isChecked()
        && !XcbUtil::get().hasState( this, XcbDefines::_NET_WM_STATE_STICKY ) )
    { _toggleShowOnAllDesktops( true ); }
    #endif
}

//__________________________________________________________
void TransparentWindow::_processDesktopChanged( WId, quint32 desktop )
{
    #if HAVE_XCB
    // make sure that on all desktop state is preserved
    if( stickyAction_->isChecked() && desktop != XcbDefines::ALL_DESKTOPS ) _toggleShowOnAllDesktops( true );
    #endif
}

//________________________________________________
void TransparentWindow::_installContextMenuActions( BaseContextMenu* menu )
{

    Debug::Throw( "TransparentWindow::_installContextMenuActions.\n" );

    menuSeparator_ = menu->addSeparator();
    menu->addAction( &_configurationAction() );

    auto application( Base::Singleton::get().application<BaseApplication>() );

    {
        QMenu* flagsMenu = menu->QMenu::addMenu( tr( "Flags" ) );
        flagsMenu->addAction( staysOnTopAction_ );
        flagsMenu->addAction( hideFromTaskBarAction_ );
        flagsMenu->addAction( stickyAction_ );
        flagsMenu->addAction( lockPositionAction_ );
        flagsMenu->addAction( roundCornersAction_ );
        flagsMenu->addAction( useSvgAction_ );
        flagsMenu->addAction( &inverseColorsAction() );
        flagsMenu->addSeparator();
        flagsMenu->addAction( lockFlagsAction_ );
    }

    {
        QMenu *helpMenu = menu->QMenu::addMenu( tr( "Help" ) );
        helpMenu->addAction( &help_->displayAction() );
        #if HAVE_XCB
        if( XcbUtil::isX11() )
        { helpMenu->addAction( &reloadBlurRegionAction() ); }
        #endif
        helpMenu->addAction( reloadBackgroundAction_ );
        helpMenu->addSeparator();

        helpMenu->addAction( &application->aboutAction() );
        helpMenu->addAction( &application->aboutQtAction() );

    }

    // debug menu
    DebugMenu *debugMenu( new DebugMenu( menu ) );
    menu->addAction( debugMenu->menuAction() );
    debugMenu->menuAction()->setVisible( false );

    menu->addSeparator();
    menu->addAction( &application->closeAction() );

    return;

}

//________________________________________________
void TransparentWindow::_configureFromDialog( ConfigurationDialog& dialog )
{
    Debug::Throw( "TransparentWindow::_configureFromDialog.\n" );
    connect( &dialog, SIGNAL(configurationChanged()), SLOT(_updateConfiguration()) );
    connect( &dialog, SIGNAL(configurationChanged()), Base::Singleton::get().application(), SIGNAL(configurationChanged()) );
    dialog.centerOnParent();
    dialog.exec();
}

//_______________________________________________________
void TransparentWindow::changeEvent( QEvent* event )
{

    if( event->type() == QEvent::PaletteChange &&
        useSvgAction_->isChecked() &&
        Svg::SvgEngine::get().isValid() &&
        Svg::SvgEngine::get().needsReloadOnPaletteChange() )
    { Svg::SvgEngine::get().forceReload(); }

    return Transparency::TransparentWidget::changeEvent( event );

}

//_______________________________________________________
void TransparentWindow::timerEvent( QTimerEvent* event )
{

    if( event->timerId() == geometryTimer_.timerId() )
    {

        // stop timer
        geometryTimer_.stop();
        _saveGeometry();

        // send geometry changed signal
        emit geometryChanged();

    } else return Transparency::TransparentWidget::timerEvent( event );

}

//___________________________________________________
void TransparentWindow::moveEvent( QMoveEvent* event )
{
    TransparentWidget::moveEvent( event );
    geometryTimer_.start( 200, this );
}

//___________________________________________________
void TransparentWindow::resizeEvent( QResizeEvent *event )
{
    Debug::Throw( "TransparentWindow::resizeEvent.\n" );
    geometryTimer_.start( 200, this );
    _updateMask();

    if( Transparency::CompositeEngine::get().isAvailable() )
    { _updateBlurRegion( _outerPadding().adjustedRect( rect() ).translated( mapTo( window(), QPoint(0,0) ) ) ); }

    TransparentWidget::resizeEvent( event );
    WinUtil( this ).enableBlurBehind( _outerPadding() );
}

//___________________________________________________
void TransparentWindow::mousePressEvent( QMouseEvent *event )
{

    Transparency::TransparentWidget::mousePressEvent( event );

    _setButton( event->button() );
    if( button_ == Qt::LeftButton && !lockPositionAction_->isChecked() )
    {
        // check modifiers
        if( event->modifiers() & Qt::ShiftModifier ) direction_ = _direction( event->pos() );
        else direction_ = XcbDefines::_NET_WM_MOVERESIZE_MOVE;

        clickPosition_ = event->pos() + QPoint(geometry().topLeft() - frameGeometry().topLeft());
        clickSize_ = size();

        #if HAVE_XCB
        // try use X11 Window Manager to handle event
        if( XcbUtil::get().moveResizeWidget( this, event->globalPos(), direction_ ) ) return;
        else
        #endif
        {

            // otherwise use curstom
            if( event->modifiers() & Qt::ShiftModifier )
            {
                setCursor( ( direction_ == XcbDefines::_NET_WM_MOVERESIZE_SIZE_TOPLEFT || direction_ == XcbDefines::_NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT ) ? Qt::SizeFDiagCursor : Qt::SizeBDiagCursor );
            } else setCursor( Qt::SizeAllCursor );

        }

    }

}

//___________________________________________________
void TransparentWindow::mouseReleaseEvent( QMouseEvent* )
{
    _setButton( Qt::NoButton );
    direction_ = XcbDefines::_NET_WM_MOVERESIZE_MOVE;
    unsetCursor();
}

//___________________________________________________
void TransparentWindow::mouseMoveEvent( QMouseEvent *event )
{

    if( button_ == Qt::LeftButton && !lockPositionAction_->isChecked() )
    {

        QPoint globalPoint( event->globalPos() - clickPosition_ );
        if( direction_ == XcbDefines::_NET_WM_MOVERESIZE_MOVE ) move( globalPoint );
        else {

            QRect newRect( geometry() );
            QPoint point( event->pos() - clickPosition_ );
            if( direction_ == XcbDefines::_NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT )
            {

                newRect.setWidth( clickSize_.width() + point.x() );
                newRect.setHeight( clickSize_.height() + point.y() );

            } else if( direction_ == XcbDefines::_NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT ) {

                newRect.setLeft( globalPoint.x() );
                newRect.setHeight( clickSize_.height() + point.y() );

            } else if( direction_ == XcbDefines::_NET_WM_MOVERESIZE_SIZE_TOPRIGHT ) {

                newRect.setWidth( clickSize_.width() + point.x() );
                newRect.setTop( globalPoint.y() );

            } else {

                newRect.setLeft( globalPoint.x() );
                newRect.setTop( globalPoint.y() );

            }

            setGeometry( newRect );

        }

    }
}

//___________________________________________________
void TransparentWindow::mouseDoubleClickEvent( QMouseEvent *event )
{
    Transparency::TransparentWidget::mouseDoubleClickEvent( event );
    _setButton( event->button() );
    return;
}

//________________________________________________
void TransparentWindow::contextMenuEvent( QContextMenuEvent* event )
{

    Debug::Throw( "TransparentWindow::contextMenuEvent.\n" );

    BaseContextMenu menu( this );
    menu.setHideDisabledActions( true );
    _installContextMenuActions( &menu );
    menu.exec( event->globalPos() );

}


//________________________________________________
void TransparentWindow::_configuration()
{

    // save configuration
    Debug::Throw( "TransparentWindow::_configuration.\n" );
    ConfigurationDialog dialog;
    _configureFromDialog( dialog );

}

//________________________________________________
void TransparentWindow::_updateConfiguration()
{

    Debug::Throw( "TransparentWindow::_updateConfiguration.\n" );

    // actions
    staysOnTopAction_->setChecked( XmlOptions::get().get<bool>( "STAYS_ON_TOP" ) );
    lockPositionAction_->setChecked( XmlOptions::get().get<bool>( "LOCK_POSITION" ) );
    roundCornersAction_->setChecked( XmlOptions::get().get<bool>( "ROUND_CORNERS" ) );

    // otherwise toggle the action directly
    stickyAction_->setChecked( XmlOptions::get().get<bool>( "STICKY" ) );
    hideFromTaskBarAction_->setChecked( XmlOptions::get().get<bool>( "HIDE_FROM_TASKBAR" ) );

    // reload svg
    useSvgAction_->setChecked( XmlOptions::get().get<bool>( "USE_SVG" ) );

    if( useSvgAction_->isChecked() && Svg::SvgEngine::get().reload() )
    { setBackgroundChanged( true ); }

    // reload background
    #if !HAVE_XCB
    reloadBackgroundAction_->setEnabled( useSvgAction_->isChecked() );
    #endif

    // flags locking
    lockFlagsAction_->setChecked( XmlOptions::get().get<bool>( "LOCK_FLAGS" ) );

    _updateMargins();

}

//________________________________________________
void TransparentWindow::_updateMargins()
{

    Debug::Throw( "TransparentWindow::_updateMargins.\n" );

    // margins
    auto margins = Base::Margins( XmlOptions::get().get<int>( "WINDOW_BORDER" ) );
    if( useSvgAction_->isChecked() && Svg::SvgEngine::get().isValid() )
    {
        margins += Svg::SvgEngine::get().margins();
        margins -= Base::Margins( 4 );
    }

    _setMargins( margins );

    // outer padding
    _clearOuterPadding();
    if( useSvgAction_->isChecked() && Svg::SvgEngine::get().isValid() )
    {
        // read either outer padding or margins from SVG, and adjust
        auto outerPadding = Svg::SvgEngine::get().outerPadding();
        if( outerPadding.isNull() )
        {
            outerPadding = Svg::SvgEngine::get().margins();
            if( !outerPadding.isNull() ) outerPadding -= Base::Margins( 4 );
        }

        _setOuterPadding( outerPadding );

    }

    // update geometry
    _loadGeometry();
}

//________________________________________________
void TransparentWindow::_toggleStaysOnTop( bool state )
{
    Debug::Throw( "TransparentWindow::_toggleStaysOnTop.\n" );

    // check that widget is top level
    if( parentWidget() ) return;

    #if HAVE_XCB

    // on X11 one uses NET_WM directly
    if( XcbUtil::get().isSupported( XcbDefines::_NET_WM_STATE_STAYS_ON_TOP ) ) XcbUtil::get().changeState( this, XcbDefines::_NET_WM_STATE_STAYS_ON_TOP, state );
    if( XcbUtil::get().isSupported( XcbDefines::_NET_WM_STATE_ABOVE ) ) XcbUtil::get().changeState( this, XcbDefines::_NET_WM_STATE_ABOVE, state );

    #else

    bool visible( !isHidden() );
    if( visible )
    {
        hide();
        qApp->processEvents();
    }

    // Qt implementation
    if( state ) setWindowFlags( windowFlags() | Qt::WindowStaysOnTopHint );
    else setWindowFlags( windowFlags() & ~Qt::WindowStaysOnTopHint );

    if( visible ) show();

    #endif

    XmlOptions::get().set<bool>( "STAYS_ON_TOP", state );
    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "STAYS_ON_TOP", XmlOptions::get().option( "STAYS_ON_TOP" ) ) ); }

}

//________________________________________________
void TransparentWindow::_toggleLockPosition( bool state )
{

    XmlOptions::get().set<bool>( "LOCK_POSITION",  state );
    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "LOCK_POSITION", XmlOptions::get().option( "LOCK_POSITION" ) ) ); }

}

//________________________________________________
void TransparentWindow::_toggleHideFromTaskBar( bool state )
{

    Debug::Throw() << "TransparentWindow::_toggleHideFromTaskBar - state: " << state << endl;

    // check that widget is top level
    if( parentWidget() ) return;

    #if HAVE_XCB
    if(  XcbUtil::get().isSupported( XcbDefines::_NET_WM_STATE_SKIP_TASKBAR ) ) XcbUtil::get().changeState( this, XcbDefines::_NET_WM_STATE_SKIP_TASKBAR, state );
    if(  XcbUtil::get().isSupported( XcbDefines::_NET_WM_STATE_SKIP_PAGER ) ) XcbUtil::get().changeState( this, XcbDefines::_NET_WM_STATE_SKIP_PAGER, state );
    #endif

    #if defined(Q_OS_WIN)
    WinUtil( this ).toggleHideFromTaskBar( state );
    #endif

    // update option
    XmlOptions::get().set<bool>( "HIDE_FROM_TASKBAR", state );
    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "HIDE_FROM_TASKBAR", XmlOptions::get().option( "HIDE_FROM_TASKBAR" ) ) ); }

    return;

}

//________________________________________________
void TransparentWindow::_toggleShowOnAllDesktops( bool state )
{

    Debug::Throw() << "TransparentWindow::_toggleShowOnAllDesktops - state: " << state << endl;

    // check that widget is top level
    if( parentWidget() ) return;

    #if HAVE_XCB
    if( XcbUtil::get().isSupported( XcbDefines::_NET_WM_STATE_STICKY ) )
    {

        // change property depending on state
        XcbUtil::get().changeState( this, XcbDefines::_NET_WM_STATE_STICKY, state );

    } else if( XcbUtil::get().isSupported( XcbDefines::_NET_WM_DESKTOP ) ) {

        if( state ) XcbUtil::get().changeCardinal( this, XcbDefines::_NET_WM_DESKTOP, XcbDefines::ALL_DESKTOPS );
        else if( XcbUtil::get().isSupported( XcbDefines::_NET_CURRENT_DESKTOP ) ) {

            auto desktop = XcbUtil::get().cardinal( XcbUtil::get().appRootWindow(), XcbDefines::_NET_CURRENT_DESKTOP );
            XcbUtil::get().changeCardinal( this, XcbDefines::_NET_WM_DESKTOP, desktop );

        }

    }

    #endif

    // update option
    XmlOptions::get().set<bool>( "STICKY", state );
    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "STICKY", XmlOptions::get().option( "STICKY" ) ) ); }

}

//________________________________________________
void TransparentWindow::_toggleRoundCorners( bool state )
{

    _updateMask();

    // only update options if action is enabled
    XmlOptions::get().set<bool>( "ROUND_CORNERS", state );
    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "ROUND_CORNERS", XmlOptions::get().option( "ROUND_CORNERS" ) ) ); }

    return;

}

//________________________________________________
void TransparentWindow::_toggleSvg( bool state )
{

    if( state && !Svg::SvgEngine::get().isValid() ) Svg::SvgEngine::get().reload();

    // only update options if action is enabled
    if( useSvgAction_->isEnabled() )
    {
        XmlOptions::get().set<bool>( "USE_SVG", state );
        if( lockFlagsAction_->isChecked() )
        { emit optionChanged( XmlOption( "USE_SVG", XmlOptions::get().option( "USE_SVG" ) ) ); }
    }

    // update background
    setBackgroundChanged( true );
    update();

    return;
}

//____________________________________________________________________
void TransparentWindow::_toggleInverseColors( bool value )
{

    Debug::Throw( "TransparentWindow::_toggleInverseColors" );
    Transparency::TransparentWidget::_toggleInverseColors( value );

    if( lockFlagsAction_->isChecked() )
    { emit optionChanged( XmlOption( "TRANSPARENCY_INVERSE_COLORS", XmlOptions::get().option( "TRANSPARENCY_INVERSE_COLORS" ) ) ); }

    update();
}

//____________________________________________________________
void TransparentWindow::_toggleLockFlags( bool state )
{

    XmlOptions::get().set<bool>( "LOCK_FLAGS", state );
    return;

}

//________________________________________________
void TransparentWindow::_updateBackgroundPixmap()
{

    TransparentWidget::_updateBackgroundPixmap();
    Debug::Throw()
        << "TransparentWindow::_updateBackgroundPixmap -"
        << " size: " << _backgroundPixmap().width() << "," << _backgroundPixmap().height()
        << endl;

    if( useSvgAction_->isChecked() && Svg::SvgEngine::get().isValid() )
    {
        QPainter painter( &_backgroundPixmap() );
        painter.setRenderHints(QPainter::SmoothPixmapTransform);
        painter.setCompositionMode(QPainter::CompositionMode_Source );
        painter.drawPixmap( QPoint( 0, 0 ), Svg::SvgEngine::get().get( Svg::SvgId( size() ) ) );
        painter.end();
    }

}

//________________________________________________
void TransparentWindow::_loadGeometryDelayed()
{
    Debug::Throw() << "TransparentWindow::_loadGeometryDelayed" << endl;
    QTimer::singleShot( 1000, this, SLOT(_loadGeometry()) );
}

//________________________________________________
void TransparentWindow::_loadGeometry()
{
    Debug::Throw() << "TransparentWindow::_loadGeometry" << endl;

    // position
    QPoint position(0, 0);
    if( XmlOptions::get().contains( "WINDOW_X" ) && XmlOptions::get().contains( "WINDOW_Y" ) )
    { position = QPoint( XmlOptions::get().get<int>( "WINDOW_X" ), XmlOptions::get().get<int>( "WINDOW_Y" ) ); }

    // size
    QSize size( 120, 120 );
    if( XmlOptions::get().contains( "WINDOW_WIDTH" ) && XmlOptions::get().contains( "WINDOW_HEIGHT" ) )
    { size = QSize( XmlOptions::get().get<int>( "WINDOW_WIDTH" ), XmlOptions::get().get<int>( "WINDOW_HEIGHT" ) ); }

    // anchor
    if( XmlOptions::get().contains( "ANCHOR" ) ) anchor_ = static_cast<Anchor::AlignmentFlags>( XmlOptions::get().get<int>( "ANCHOR" ) );
    else anchor_ = Anchor::Default;

    // order
    if( XmlOptions::get().contains( "ORDER" ) ) order_ = XmlOptions::get().get<int>( "ORDER" );
    else order_ = 0;

    // update position to account for anchor
    QDesktopWidget* desktop( qApp->desktop() );
    const QRect desktopGeometry( desktop->availableGeometry( desktop->screenNumber( this ) ) );
    {

        // horizontal positioning
        if( anchor_ & Anchor::Left ) position.setX( desktopGeometry.left() + position.x() );
        else if( anchor_ & Anchor::Right ) position.setX( desktopGeometry.right() - position.x() );
        else if( anchor_ & Anchor::HorizontalCenter ) position.setX( position.x() + desktopGeometry.center().x()  );

        // vertical positioning
        if( anchor_ & Anchor::Top ) position.setY( desktopGeometry.top() + position.y() );
        else if( anchor_ & Anchor::Bottom ) position.setY( desktopGeometry.bottom() - position.y() );
        else if( anchor_ & Anchor::VerticalCenter ) position.setY( position.y() + desktopGeometry.center().y() );

    }

    // set geometry, properly accounting for outer padding
    setInnerGeometry( QRect( position, size ) );

}

//________________________________________________
void TransparentWindow::_saveGeometry() const
{
    Debug::Throw() << "TransparentWindow::_saveGeometry" << endl;

    // get current geometry
    QRect geometry( this->geometry() );

    // take out margins from svg
    geometry = _outerPadding().adjustedRect( this->geometry() );

    // store position and size
    const QSize size( geometry.size() );
    QPoint position( geometry.topLeft() );

    // change position to account for anchor
    QDesktopWidget* desktop( qApp->desktop() );
    const QRect desktopGeometry( desktop->availableGeometry( desktop->screenNumber( this ) ) );
    {

        // horizontal positioning
        if( anchor_ & Anchor::Left ) position.setX( position.x() - desktopGeometry.left() );
        else if( anchor_ & Anchor::Right ) position.setX( desktopGeometry.right() - position.x() );
        else if( anchor_ & Anchor::HorizontalCenter ) position.setX( position.x() - desktopGeometry.center().x() );

        // vertical positioning
        if( anchor_ & Anchor::Top ) position.setY( position.y() - desktopGeometry.top() );
        if( anchor_ & Anchor::Bottom ) position.setY( desktopGeometry.bottom() - position.y() );
        else if( anchor_ & Anchor::VerticalCenter ) position.setY( position.y() - desktopGeometry.center().y() );

    }

    // save everything
    XmlOptions::get().set<int>( "WINDOW_X", position.x() );
    XmlOptions::get().set<int>( "WINDOW_Y", position.y() );

    if( size.width() > 0 ) XmlOptions::get().set<int>( "WINDOW_WIDTH", size.width() );
    if( size.height() > 0 ) XmlOptions::get().set<int>( "WINDOW_HEIGHT", size.height() );
    XmlOptions::get().set<int>( "ANCHOR", anchor_ );
    XmlOptions::get().set<int>( "ORDER", order_ );

    // write
    XmlOptions::write();
}

//________________________________________________
void TransparentWindow::_installActions()
{
    Debug::Throw( "TransparentWindow::_installActions.\n" );

    // configure
    configurationAction_ = new QAction( IconEngine::get( IconNames::Configure ), tr( "Configuration" ), this );
    configurationAction_->setToolTip( tr( "Application configuration" ) );
    connect( configurationAction_, SIGNAL(triggered()), SLOT(_configuration()) );

    // stays on top
    addAction( staysOnTopAction_ = new QAction( tr( "Keep Above" ), this ) );
    staysOnTopAction_->setToolTip( tr( "Keep window above all others" ) );
    staysOnTopAction_->setCheckable( true );
    staysOnTopAction_->setChecked( windowFlags() & Qt::WindowStaysOnTopHint );
    connect( staysOnTopAction_, SIGNAL(toggled(bool)), SLOT(_toggleStaysOnTop(bool)) );

    // lock
    addAction( lockPositionAction_ = new QAction( tr( "Lock Position" ), this ) );
    lockPositionAction_->setCheckable( true );
    lockPositionAction_->setEnabled( windowFlags() & Qt::FramelessWindowHint );
    connect( lockPositionAction_, SIGNAL(toggled(bool)), SLOT(_toggleLockPosition(bool)) );

    // hide from taskBar
    addAction( hideFromTaskBarAction_ = new QAction( tr( "Hide from TaskBar" ), this ) );
    hideFromTaskBarAction_->setCheckable( true );
    hideFromTaskBarAction_->setChecked( false );
    connect( hideFromTaskBarAction_, SIGNAL(toggled(bool)), SLOT(_toggleHideFromTaskBar(bool)) );

    // sticky
    addAction( stickyAction_ = new QAction( tr( "Show on all Desktops" ), this ) );
    stickyAction_->setToolTip( tr( "Make window appear on all desktops" ) );
    stickyAction_->setCheckable( true );
    stickyAction_->setChecked( false );
    connect( stickyAction_, SIGNAL(toggled(bool)), SLOT(_toggleShowOnAllDesktops(bool)) );

    // round corners
    addAction( roundCornersAction_ = new QAction( tr( "Round Corners" ), this ) );
    roundCornersAction_->setToolTip( tr( "Round window corners" ) );
    roundCornersAction_->setCheckable( true );
    roundCornersAction_->setChecked( false );
    roundCornersAction_->setEnabled( windowFlags() & Qt::FramelessWindowHint );
    connect( roundCornersAction_, SIGNAL(toggled(bool)), SLOT(_toggleRoundCorners(bool)) );

    // svg
    addAction( useSvgAction_ = new QAction( tr( "Use Svg" ), this ) );
    useSvgAction_->setToolTip( tr( "Use svg for background" ) );
    useSvgAction_->setCheckable( true );
    useSvgAction_->setChecked( false );
    connect( useSvgAction_, SIGNAL(toggled(bool)), SLOT(_toggleSvg(bool)) );

    // reload background
    addAction( reloadBackgroundAction_ = new QAction( IconEngine::get( IconNames::Reload ), tr( "Reload Background" ), this ) );
    reloadBackgroundAction_->setToolTip( tr( "Reload window background" ) );
    reloadBackgroundAction_->setShortcut( QKeySequence::Refresh );
    connect( reloadBackgroundAction_, SIGNAL(triggered()), &Svg::SvgEngine::get(), SLOT(forceReload()) );

    #if HAVE_XCB
    // also connect to blur background
    if( XcbUtil::isX11() )
    { connect( reloadBackgroundAction_, SIGNAL(triggered()), &reloadBlurRegionAction(), SLOT(trigger()) ); }
    #endif

    // lock
    addAction( lockFlagsAction_ = new QAction( tr( "Lock Flags" ), this ) );
    lockFlagsAction_->setToolTip( tr( "Use common flags for all transparency applications" ) );
    lockFlagsAction_->setCheckable( true );
    lockFlagsAction_->setChecked( false );
    connect( lockFlagsAction_, SIGNAL(toggled(bool)), SLOT(_toggleLockFlags(bool)) );

    #if HAVE_XCB
    hideFromTaskBarAction_->setEnabled( XcbUtil::get().isSupported( XcbDefines::_NET_WM_STATE_SKIP_TASKBAR ) );
    stickyAction_->setEnabled(
        XcbUtil::get().isSupported( XcbDefines::_NET_WM_STATE_STICKY ) ||
        XcbUtil::get().isSupported( XcbDefines::_NET_WM_DESKTOP ) );
    #endif

}

//________________________________________________
void TransparentWindow::_updateMask()
{
    if( roundCornersAction_->isChecked() ) setMask( RoundedRegion( rect() ).get() );
    else clearMask();
}

//________________________________________________
XcbDefines::Direction TransparentWindow::_direction( const QPoint& point ) const
{
    const int halfWidth( width()/2 );
    const int halfHeight( height()/2 );
    if( point.x() < halfWidth ) return point.y() < halfHeight ?  XcbDefines::_NET_WM_MOVERESIZE_SIZE_TOPLEFT: XcbDefines::_NET_WM_MOVERESIZE_SIZE_BOTTOMLEFT;
    else return point.y() < halfHeight ? XcbDefines::_NET_WM_MOVERESIZE_SIZE_TOPRIGHT: XcbDefines::_NET_WM_MOVERESIZE_SIZE_BOTTOMRIGHT;
}

//__________________________________________________________
void TransparentWindow::Color::update( const QColor& color )
{

    Debug::Throw( "TransparentWindow::Color::update.\n" );
    if( !color.isValid() ) return;

    ( foreground_ = color ).setAlpha( XmlOptions::get().get<int>( "TRANSPARENCY_FOREGROUND_INTENSITY" ) );
    ( background_ = color ).setAlpha( XmlOptions::get().get<int>( "TRANSPARENCY_BACKGROUND_INTENSITY" ) );

    // update gradient
    const int lightenAmount( 130 );
    useGradient_ = XmlOptions::get().get<bool>( "TRANSPARENCY_USE_GRADIENT" );
    foregroundGradient_ = QLinearGradient( foregroundGradient_.start(), foregroundGradient_.finalStop() );
    foregroundGradient_.setColorAt(0, foreground_.lighter( lightenAmount ) );
    foregroundGradient_.setColorAt(1, foreground_ );

    backgroundGradient_ = QLinearGradient( backgroundGradient_.start(), backgroundGradient_.finalStop() );
    backgroundGradient_.setColorAt(0, background_.lighter( lightenAmount ) );
    backgroundGradient_.setColorAt(1, background_ );

}

//__________________________________________________________
void TransparentWindow::Color::updateGradient( const QSize& size )
{
    Debug::Throw( "TransparentWindow::Color::updateGradient.\n" );
    foregroundGradient_.setFinalStop( 0, size.height() );
    backgroundGradient_.setFinalStop( 0, size.height() );
}

//__________________________________________________________
QIcon TransparentWindow::Color::icon() const
{
    Debug::Throw( "TransparentWindow::Color::icon.\n" );

    QPixmap pixmap( CustomPixmap( IconSize::get( IconSize::Large ), CustomPixmap::Flag::All ) );
    QPainter painter( &pixmap );
    painter.setRenderHints(QPainter::Antialiasing );
    painter.setPen( foreground_ );
    painter.setBrush( background_ );

    QRectF rect = QRect( QPoint(0,0), IconSize::get( IconSize::Large ) ).adjusted( 2, 2, -2, -2 );
    painter.drawEllipse( rect );
    painter.end();

    return QIcon( pixmap );
}
