#ifndef Temperature_h
#define Temperature_h

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "Counter.h"
#include "CustomProcess.h"
#include "MessageBuffer.h"

#include <QObject>
#include <QHash>
#include <QList>
#include <QProcess>
#include <QSet>
#include <QString>
#include <QTextStream>

namespace Temperature
{

    //* measure relative cpu load from /proc/stat file
    class Temperature: public QObject, private Base::Counter<Temperature>
    {

        Q_OBJECT

        public:

        //* constructor
        explicit Temperature();

        //* use sensors
        bool useSensors() const
        { return useSensors_; }

        //* use sensors
        void setUseSensors( bool value )
        { useSensors_ = value; }


        //* temperature data
        class Data
        {

            public:

            //* constructor
            explicit Data( double temperature = 0, bool valid = false ):
                temperature_( temperature ),
                valid_( valid )
            {}

            //* load from stream
            friend QTextStream &operator >> ( QTextStream &in, Data& data )
            {
                in >> data.temperature_;
                data.valid_ = true;

                return in;
            }

            //* sum points
            Data& operator += ( const Data& other )
            {
                temperature_ += other.temperature_;
                valid_ &= other.valid_;
                return *this;
            }

            //* sum points
            Data& operator -= ( const Data& other )
            {
                temperature_ -= other.temperature_;
                valid_ &= other.valid_;
                return *this;
            }

            //* divide point by an integer
            Data& operator /= (const unsigned long& count )
            {
                temperature_ /= count;
                return *this;
            }

            //* temperature
            double temperature() const
            { return temperature_; }

            //* validity
            bool valid() const
            { return valid_; }

            private:

            //* temperature
            double temperature_ = 0;

            //* true if current data are valid
            bool valid_ = false;

        };

        //* shortcut for data list
        using DataList = QList< Data >;

        //* Temperature data samples storage
        class Samples
        {

            public:

            //* construtor
            explicit Samples( int sampleSize = 1, int dataSize = 10 ):
                sampleSize_( sampleSize ),
                dataSize_( dataSize )
            {}

            //* set sample size
            void setSampleSize( const int sampleSize );

            //* set data point size
            void setDataSize( const int dataSize );

            //* retrieve data
            const DataList& data() const
            { return dataList_; }

            //* add a data point
            void add( const Data& data );

            private:

            //* list of cpu load measurement
            DataList sampleList_;

            //* list of cpu load differences
            DataList dataList_;

            //* number of samples to be averaged to make the data points
            int sampleSize_ = 1;

            //* number of data point to store
            int dataSize_ = 10;

        };

        //* shortcut for Samples set
        using SampleMap = QHash< QString, Samples >;

        //* cpu names
        using ProbeNames = QSet<QString>;

        //* get available cpu names
        ProbeNames probeNames() const
        {
            ProbeNames out;
            for( SampleMap::const_iterator iter = sampleMap_.begin(); iter != sampleMap_.end(); iter++ )
            { out.insert( iter.key() ); }
            return out;
        }

        //* changes sampleSize_
        void setSampleSize( int value )
        {
            sampleSize_ = std::max<int>( value, 1 );
            for( SampleMap::iterator iter = sampleMap_.begin(); iter != sampleMap_.end(); iter++ )
            { iter.value().setSampleSize( value ); }
        }

        //* changes sampleSize_
        void setDataSize( int value )
        {
            dataSize_ = std::max<int>( value, 1 );
            for( SampleMap::iterator iter = sampleMap_.begin(); iter != sampleMap_.end(); iter++ )
            { iter.value().setDataSize( value ); }
        }

        //* retrieve data for a given cpu name
        DataList data( const QString& ) const;

        //* update all dataSamples from file
        void update();

        protected Q_SLOTS:

        //* take action when current command completed
        void _commandCompleted( int, QProcess::ExitStatus );

        //* read on-fly from process
        void _readStdout();

        //* read on-fly from process
        void _readStderr();

        private:

        //* true if use lm sensors
        bool useSensors_ = true;

        //* custom process
        /** used to monitor temperature using LM sensors */
        CustomProcess process_;

        //* stdout buffer
        MessageBuffer stdoutBuffer_;

        //* stderr buffer
        MessageBuffer stderrBuffer_;

        //* number of samples to average on
        int sampleSize_ = 1;

        //* number of data to be kept
        int dataSize_ = 10;

        //* data samples
        SampleMap sampleMap_;

    };

    //* less than operator
    inline bool operator < (const Temperature::Data& first, const Temperature::Data& second)
    { return first.temperature() < second.temperature(); }

};

#endif
