/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TemperatureConfigurationDialog.h"

#include "GridLayout.h"
#include "IconEngine.h"
#include "IconNames.h"
#include "OptionCheckBox.h"
#include "OptionColorDisplay.h"
#include "OptionSlider.h"
#include "OptionSpinBox.h"
#include "OptionBrowsedLineEditor.h"

#include <QGroupBox>
#include <QLabel>
#include <QLayout>
#include <QToolButton>

namespace Temperature
{

    //_________________________________________________________
    TemperatureConfigurationDialog::TemperatureConfigurationDialog( QWidget* parent ):
        ConfigurationDialog( parent )
    {

        Debug::Throw( "TemperatureConfigurationDialog::TemperatureConfigurationDialog.\n" );
        setWindowTitle( "Configuration - Temperature Monitor" );

        // generic widgets

        QWidget* page;
        page = &addPage( IconEngine::get( IconNames::PreferencesAppearance), tr( "Appearance" ), tr( "Window appearance and colors" ) );
        {

            QWidget *box;
            page->layout()->addWidget( box = new QWidget( page ));

            QVBoxLayout* vLayout( new QVBoxLayout );
            vLayout->setSpacing(5);
            vLayout->setMargin(0);
            box->setLayout( vLayout );

            // use default colors
            OptionCheckBox* checkbox;
            vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Use default colors" ), box, "USE_DEFAULT_COLORS" ) );
            addOptionWidget( checkbox );

            box = new QWidget( box );
            vLayout->addWidget( box );

            auto gridLayout = new GridLayout;
            gridLayout->setSpacing(5);
            gridLayout->setMargin(0);
            gridLayout->setMaxCount( 2 );
            gridLayout->setColumnAlignment( 0, Qt::AlignRight|Qt::AlignVCenter );
            box->setLayout( gridLayout );

            // temperature color
            QLabel* label;
            gridLayout->addWidget( label = new QLabel( tr( "Color:" ), box ) );

            OptionColorDisplay *colorDisplay;
            gridLayout->addWidget( colorDisplay = new OptionColorDisplay( box, "TEMPERATURE_COLOR" ) );
            addOptionWidget( colorDisplay );
            label->setBuddy( colorDisplay );

            connect( checkbox, SIGNAL(toggled(bool)), label, SLOT(setDisabled(bool)) );
            connect( checkbox, SIGNAL(toggled(bool)), colorDisplay, SLOT(setDisabled(bool)) );

            // intensity
            gridLayout->addWidget( label = new QLabel( tr( "Intensity:" ), this ) );

            OptionSlider* slider;
            gridLayout->addWidget( slider = new OptionSlider( this, "TRANSPARENCY_BACKGROUND_INTENSITY" ) );
            slider->setScale( 1.0/2.55 );
            slider->setRange( 0, 100 );
            addOptionWidget( slider );
            label->setBuddy( slider );

            vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Use gradient" ), box, "TRANSPARENCY_USE_GRADIENT" ) );
            addOptionWidget( checkbox );

            vLayout->addWidget( checkbox = new OptionCheckBox( tr( "Show temperature" ), box, "SHOW_LOAD" ) );
            checkbox->setToolTip( tr( "Show/hide cpu load percentage" ) );
            addOptionWidget( checkbox );

            vLayout->addWidget( checkbox = new OptionCheckBox(  tr( "Show hostname" ), box, "SHOW_HOST" ) );
            checkbox->setToolTip( tr( "Show/hide host name" ) );
            addOptionWidget( checkbox );
        }

        // Unsorted
        page = &addPage( IconEngine::get( IconNames::PreferencesUnsorted ), tr( "Unsorted" ), tr( "Additional unsorted settings" ) );
        {

            QWidget *box;
            page->layout()->addWidget( box = new QWidget( page ) );
            auto gridLayout = new QGridLayout;
            gridLayout->setSpacing(5);
            gridLayout->setMargin(0);
            box->setLayout( gridLayout );

            OptionCheckBox* checkbox;
            gridLayout->addWidget( checkbox = new OptionCheckBox(  tr( "Use lm-sensors" ), box, "USE_LM_SENSORS" ), 0, 1, 1, 2 );
            checkbox->setToolTip( tr( "Use lm-sensors to get temperatures" ) );
            checkbox->setChecked( true );
            addOptionWidget( checkbox );

            QLabel* label;
            gridLayout->addWidget( label = new QLabel( tr( "Lm-sensors command:" ), box ), 1, 0, 1, 1 );

            OptionBrowsedLineEditor* editor;
            gridLayout->addWidget( editor = new OptionBrowsedLineEditor( box, "LM_SENSORS" ), 1, 1, 1, 1 );
            addOptionWidget( editor );
            label->setBuddy( editor );
            label->setAlignment(Qt::AlignVCenter|Qt::AlignRight);
            connect( checkbox, SIGNAL(toggled(bool)), editor, SLOT(setEnabled(bool)) );

            editor->setTargetApplication( File("sensors") );

            gridLayout->addWidget( label = new QLabel( tr( "ACPI file:" ), box ), 2, 0, 1, 1 );
            gridLayout->addWidget( editor = new OptionBrowsedLineEditor( box, "ACPI_FILE" ), 2, 1, 1, 2 );
            addOptionWidget( editor );
            editor->setEnabled( false );
            label->setAlignment(Qt::AlignVCenter|Qt::AlignRight);
            label->setBuddy( editor );
            connect( checkbox, SIGNAL(toggled(bool)), editor, SLOT(setDisabled(bool)) );

            OptionSpinBox* spinbox;
            gridLayout->addWidget( label = new QLabel( tr( "Refresh interval:" ), box ), 3, 0, 1, 1 );
            gridLayout->addWidget( spinbox = new OptionSpinBox( box, "REFRESH_RATE" ), 3, 1, 1, 2 );
            spinbox->setSuffix( tr( "s" ) );
            spinbox->setMinimum( 1 );
            spinbox->setToolTip( tr( "Time interval (seconds) between two refresh" ) );
            addOptionWidget( spinbox );
            label->setBuddy( spinbox );
            label->setAlignment(Qt::AlignVCenter|Qt::AlignRight);

            gridLayout->addWidget( label = new QLabel( tr( "Number of samples:" ), box ), 4, 0, 1, 1 );
            gridLayout->addWidget( spinbox = new OptionSpinBox( box, "SAMPLES" ), 4, 1, 1, 2 );
            spinbox->setMinimum( 1 );
            spinbox->setToolTip( tr( "Number of samples averaged for each measurement" ) );
            addOptionWidget( spinbox );
            label->setBuddy( spinbox );
            label->setAlignment(Qt::AlignVCenter|Qt::AlignRight);

        }

        read();

    }

}
