/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "TemperatureWidget.h"

#include "BaseContextMenu.h"
#include "Color.h"
#include "InformationDialog.h"
#include "TemperatureConfigurationDialog.h"
#include "Util.h"
#include "XmlOptions.h"

#include <algorithm>

namespace Temperature
{

    //__________________________________________________________
    TemperatureWidget::TemperatureWidget( QWidget *parent ):
        HistogramWidget( parent ),
        temperatureMonitor_( this ),
        temperatureMenuAction_( 0 )
    {

        Debug::Throw( "TemperatureWidget::TemperatureWidget.\n" );
        setWindowTitle( Util::windowTitle( tr( "Temperature Monitor" ) ) );
        _configurationAction().setText( tr( "Temperature Monitor Settings..." ) );
        _updateConfiguration();
        _showLoadAction().setText( "&Show temperature" );
    }

    //______________________________________________________________________
    void TemperatureWidget::timerEvent( QTimerEvent* event )
    {

        if( event->timerId() == _updateTimer().timerId() )
        {

            temperatureMonitor_.update();
            update();

        } else HistogramWidget::timerEvent( event );

        return;

    }

    //______________________________________________________________________
    void TemperatureWidget::_paint( QPaintDevice& device, const QRect& clip )
    {

        Debug::Throw( "TemperatureWidget::_paint.\n" );

        // check margins against window size
        if( !_margins().adjustedRect( rect() ).isValid() )
        {
            InformationDialog( this, tr( "Window size is too small with respect to margins.\n Margins are set to 0." ) ).exec();
            _setMargins( 0 );
        }

        // store size
        QRect rect( _margins().adjustedRect( this->rect() ) );
        const int width( rect.width() );
        const int height( rect.height() );

        // create painter
        QPainter painter( &device );
        painter.setClipRect( clip );
        painter.setRenderHints(QPainter::Antialiasing);

        // retrieve data
        Temperature::DataList dataList( temperatureMonitor_.data() );
        while( dataList.size() > width ) dataList.removeFirst();

        if( !dataList.empty() )
        {

            // temperature
            double x( _margins().left() + 1 );
            double y_min( height + _margins().bottom() + 1 );
            double y_max( y_min );

            QPolygonF temperature_polygon;
            temperature_polygon << QPointF( x, y_min );

            // autoscale painting
            double scale( _getScale( dataList ) );
            for( const auto& data:dataList )
            {

                y_max = y_min - qMax( 1.0, scale*data.temperature() );
                temperature_polygon << QPointF( x, y_max );
                x+=1;
            }

            // add last two points to close polygons
            temperature_polygon << QPointF( x-1, y_min );
            temperature_polygon << temperature_polygon.front();

            QPen pen;
            pen.setJoinStyle( Qt::RoundJoin );
            pen.setWidthF( 1.5 );
            pen.setColor( temperatureColor_.foreground() );
            painter.setPen( pen );
            painter.setBrush( temperatureColor_.background() );
            painter.drawPolygon( temperature_polygon );

        }

        // draw host
        if( _showLoadAction().isChecked() || dataList.empty() )
        { _paintLoad( painter, rect ); }

        _paintHost( painter, rect );

        painter.end();
    }

    //______________________________________________________________________
    void TemperatureWidget::resizeEvent( QResizeEvent* event )
    {

        int width = qMax<int>( 0, event->size().width() - _margins().width() );
        temperatureMonitor_.setDataSize( width );
        temperatureColor_.updateGradient( event->size() );
        return HistogramWidget::resizeEvent( event );
    }

    //______________________________________________________________________
    void TemperatureWidget::_installContextMenuActions( BaseContextMenu* menu )
    {
        Debug::Throw( "TemperatureWidget::_installContextMenuActions.\n" );

        // create parent menu
        HistogramWidget::_installContextMenuActions( menu );

        // temperature menu
        temperatureMenuAction_ = menu->insertMenu( &_menuSeparator(), &temperatureMonitor_.menu() );

    }

    //________________________________________________
    void TemperatureWidget::_configuration()
    {

        Debug::Throw( "TemperatureWidget::_configuration.\n" );

        TemperatureConfigurationDialog dialog;
        _configureFromDialog( dialog );

    }

    //__________________________________________________________
    void TemperatureWidget::_updateConfiguration()
    {
        Debug::Throw( "TemperatureWidget::_updateConfiguration.\n" );

        // colors
        _setDefaultColors();
        if( !XmlOptions::get().get<bool>( "USE_DEFAULT_COLORS" ) )
        { temperatureColor_.update( XmlOptions::get().get<Base::Color>( "TEMPERATURE_COLOR" ) ); }

        if( temperatureMenuAction_ )
        { temperatureMenuAction_->setVisible( temperatureMonitor_.useSensors() ); }

        // base class update
        HistogramWidget::_updateConfiguration();
        update();
    }

    //____________________________________________________________________
    void TemperatureWidget::_toggleInverseColors( bool value )
    {

        Debug::Throw( "TemperatureWidget::_toggleInverseColors" );
        MainWindow::_toggleInverseColors( value );

        // update internal colors
        if( XmlOptions::get().get<bool>( "USE_DEFAULT_COLORS" ) )
        { _setDefaultColors(); }

        update();
    }

    //_________________________________________________________
    double TemperatureWidget::_getScale( const Temperature::DataList& data ) const
    {
        Debug::Throw( "TemperatureWidget::_getScale.\n" );
        Q_ASSERT( !data.empty() );

        // get maximum
        auto iter( std::max_element( data.begin(), data.end() ) );
        double maximum( iter->temperature() );

        int height( qMax<int>( 1, TemperatureWidget::height() - _margins().height() ) );
        return (maximum) ? double( height )/maximum:1;
    }

    //_________________________________________________________
    void TemperatureWidget::_paintLoad( QPainter& painter, QRect& rect )
    {
        Debug::Throw( "TemperatureWidget::_paintLoad.\n" );

        // prepare load text
        QString buffer;
        const Temperature::DataList& data( temperatureMonitor_.data() );
        if( !data.empty() ) buffer = QString( QString::fromUtf8( "%1 °C" ) ).arg( data.back().temperature() );
        else buffer = QString( tr( "Acquiring temperature" ) );

        _paintText( painter, buffer, rect );
        return;

    }

    //_________________________________________________________
    void TemperatureWidget::_setDefaultColors()
    {
        Debug::Throw() << "CpuWidget::_setDefaultColor" << endl;
        temperatureColor_.update( Base::Color( foregroundColor() ).merge( shadowColor(), 0.75 ) );
    }

}
